import pathlib
from typing import Tuple, Union

from PyQt5 import QtGui, QtWidgets

from labbie import utils

_OBSERVED_ABOUTTOQUIT_SIGNAL = False


def asset_path(subpath: Union[str, pathlib.Path]):
    return utils.assets_dir() / subpath


def fix_taskbar_icon():
    import ctypes
    myappid = 'POE-LabHelper Powered By RebrandSoft'  # arbitrary string
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)


def recolored_icon(asset, rgb: Union[int, Tuple[int, int, int]]):
    img = QtGui.QPixmap(str(asset_path(asset)))
    qp = QtGui.QPainter(img)
    qp.setCompositionMode(QtGui.QPainter.CompositionMode_SourceIn)
    if isinstance(rgb, int):
        rgb = (rgb, rgb, rgb)
    qp.fillRect(img.rect(), QtGui.QColor.fromRgb(*rgb))
    qp.end()
    return QtGui.QIcon(img)

# NOTE: hack to work around aboutToQuit firing at app initialization
def _exit_handler_wrapper(handler):
    def wrapped():
        global _OBSERVED_ABOUTTOQUIT_SIGNAL
        if not _OBSERVED_ABOUTTOQUIT_SIGNAL:
            _OBSERVED_ABOUTTOQUIT_SIGNAL = True
        else:
            handler()
    return wrapped


def register_exit_handler(handler):
    if not _OBSERVED_ABOUTTOQUIT_SIGNAL:
        handler = _exit_handler_wrapper(handler)
    QtWidgets.QApplication.instance().aboutToQuit.connect(handler)


class CheckboxProperty:
    def __init__(self, widget_fn, tristate=False):
        self.widget_fn = widget_fn
        self.tristate = tristate

    def __get__(self, obj, objtype=None):
        widget = self.widget_fn(obj)
        if not self.tristate:
            return widget.isChecked()
        else:
            return widget.checkState()

    def __set__(self, obj, value):
        widget = self.widget_fn(obj)
        if not self.tristate:
            widget.setChecked(value)
        else:
            widget.setCheckState(value)


def checkbox_property(widget_fn=None, tristate=None):
    if isinstance(tristate, bool):
        def decorator(widget_fn):
            return CheckboxProperty(widget_fn, tristate=tristate)
        return decorator
    else:
        return CheckboxProperty(widget_fn)


class RadioProperty:
    def __init__(self, widget_fn):
        self.widget_fn = widget_fn

    def __get__(self, obj, objtype=None):
        widget = self.widget_fn(obj)
        return widget.checkedId()

    def __set__(self, obj, value):
        if isinstance(value, Enum):
            value = value.value

        widget = self.widget_fn(obj)
        for btn in widget.buttons():
            if widget.id(btn) == value:
                btn.setChecked(True)


def radio_property(widget_fn):
    return RadioProperty(widget_fn)


class TextProperty:
    def __init__(self, widget_fn):
        self.widget_fn = widget_fn

    def __get__(self, obj, objtype=None):
        widget = self.widget_fn(obj)
        return widget.text()

    def __set__(self, obj, value):
        widget = self.widget_fn(obj)
        widget.setText(value)


def text_property(widget_fn):
    return TextProperty(widget_fn)


class ComboBoxProperty:
    def __init__(self, widget_fn):
        self.widget_fn = widget_fn

    def __get__(self, obj, objtype=None):
        widget = self.widget_fn(obj)
        return widget.currentText()

    def __set__(self, obj, value):
        widget = self.widget_fn(obj)
        # TODO (bnorick): use setCurrentText?
        idx = widget.findText(value)
        widget.setCurrentIndex(idx)


def combo_box_property(widget_fn):
    return ComboBoxProperty(widget_fn)


class CheckableComboBoxProperty:
    def __init__(self, widget_fn):
        self.widget_fn = widget_fn

    def __get__(self, obj, objtype=None):
        widget = self.widget_fn(obj)
        return widget.currentData()

    def __set__(self, obj, value):
        widget = self.widget_fn(obj)
        widget.setCheckedTexts(value)


def checkable_combo_box_property(widget_fn):
    return CheckableComboBoxProperty(widget_fn)



import keyboard
from PyQt5 import QtCore


class Hotkey(QtCore.QObject):
    pressed = QtCore.pyqtSignal()

    def __init__(self, hotkey):
        super().__init__()
        self.hotkey = hotkey

    def start(self, handler=None):
        keyboard.add_hotkey(self.hotkey, self.pressed.emit, suppress=True)
        if handler:
            self.pressed.connect(handler)

    def stop(self, disconnect=True):
        if disconnect:
            self.pressed.disconnect()
        keyboard.remove_hotkey(self.hotkey)

    def set_hotkey(self, hotkey):
        if self.hotkey == hotkey:
            return

        self.stop(disconnect=False)
        self.start()



import dataclasses
import string
import functools
from typing import Dict, List, Optional, Tuple, Union

import injector
import loguru
import datrie

from labbie import resources
from labbie import trade

logger = loguru.logger


@dataclasses.dataclass
class HelmModInfo:
    display: bool
    mod: str
    trade_text: Optional[str]
    trade_stat_id: Optional[str]
    trade_stat_value: Union[None, int, float]


@injector.singleton
class Mods:

    @injector.inject
    def __init__(self, resource_manager: resources.ResourceManager, trade_: trade.Trade):
        self._resource_manager = resource_manager
        self._trade = trade_

        self._raw_mods = resource_manager.mods

        self.helm_mod_info = self._build_helm_mod_info(self._raw_mods['helmet'])  # exact mod -> HelmModInfo

    @functools.cached_property
    def helm_mods(self):
        return sorted(self.helm_mod_info.keys())

    @functools.cached_property
    def helm_display_mods(self):
        return sorted(k for k, v in self.helm_mod_info.items() if v.display)

    def _build_helm_mod_info(
        self,
        mods: List[List[Tuple[str, List[str], List[Union[float, int, str]], bool]]]
    ) -> Dict[str, HelmModInfo]:
        result = {}
        for index, mod_variants in enumerate(mods):
            trade_text = None

            for mod_format, slot_patterns, _, _ in mod_variants:
                if '{' not in mod_format:
                    candidate = mod_format.lower()
                    if candidate in self._trade.text_to_stat_id:
                        trade_text = candidate
                        break

                slotted = mod_format.format(*slot_patterns).lower()
                hash_only_slot_patterns = ['#'] * mod_format.count('{')
                hash_only_slotted = mod_format.format(*hash_only_slot_patterns).lower()
                if slotted in self._trade.text_to_stat_id:
                    trade_text = slotted
                    break
                elif hash_only_slotted in self._trade.text_to_stat_id:
                    trade_text = hash_only_slotted
                    break
                else:
                    continue
            else:
                logger.warning(f'No trade text found for {index=} {mod_variants=}')

            trade_stat_id = self._trade.text_to_stat_id.get(trade_text)
            for mod_format, slot_patterns, values, display in mod_variants:
                if '{' not in mod_format:
                    result[mod_format] = HelmModInfo(
                        display=display,
                        mod=mod_format,
                        trade_text=trade_text,
                        trade_stat_id=trade_stat_id,
                        trade_stat_value=None
                    )
                    continue

                # if # ocurrs in mod_format we need to rethink some logic below
                assert '#' not in mod_format

                if len(values) == 1:
                    trade_stat_value = values[0]
                elif len(values) == 2:
                    trade_stat_value = sum(values) / 2
                else:
                    raise ValueError(f'{mod_format=} has an unexpected number of values - '
                                     f'{len(values)=} {values=}')

                slot_values = [pattern.replace('#', str(value))
                               for pattern, value in zip(slot_patterns, values)]
                mod = mod_format.format(*slot_values)
                result[mod] = HelmModInfo(
                    display=display,
                    mod=mod,
                    trade_text=trade_text,
                    trade_stat_id=trade_stat_id,
                    trade_stat_value=trade_stat_value
                )

        return result

    @functools.cached_property
    def mod_trie(self) -> datrie.Trie:
        logger.debug('Creating Trie')
        trie = datrie.Trie(string.printable)
        for mod in self.helm_mods:
            trie[mod] = 1
        return trie

    def get_mod_list_from_ocr_results(self, enchant_list):
        trie = self.mod_trie
        logger.debug(f'Data from OCR{enchant_list}')

        state = datrie.State(trie)
        parts = []
        enchants = []
        for part in enchant_list:
            to_walk = f' {part}' if parts else part
            if state.walk(to_walk):
                parts.append(to_walk)
                continue

            if not parts:
                state.rewind()
                continue

            if parts:
                enchant = ''.join(parts)
                if enchant in trie:
                    enchants.append(enchant)
                elif len(keys := trie.keys(enchant)) == 1 and len(enchant) > 20:
                    enchants.append(keys[0])
                parts = []

            state.rewind()
            if state.walk(part):
                parts.append(part)
                continue

        if parts:
            enchant = ''.join(parts)
            if enchant in trie:
                enchants.append(enchant)
            elif len(keys := trie.keys(enchant)) == 1 and len(enchant) > 20:
                enchants.append(keys[0])

        return enchants



import dataclasses

from labbie import mixins


@dataclasses.dataclass
class Bounds(mixins.SerializableMixin):
    left: int
    top: int
    right: int
    bottom: int

    def as_tuple(self):
        return (self.left, self.top, self.right, self.bottom)



class EnchantDataNotFound(Exception):
    def __init__(self, error=None):
        if error is None:
            error = 'No enchant data found.'
        super().__init__(error)


class EnchantDataInvalid(Exception):
    def __init__(self, error=None):
        if error is None:
            error = 'Invalid enchant data.'
        super().__init__(error)


class EnchantsNotLoaded(Exception):
    def __init__(self, error=None):
        if error is None:
            error = 'Enchants are not loaded.'
        super().__init__(error)


class NoSuchEnchant(Exception):
    def __init__(self, error=None):
        if error is None:
            error = 'No matching enchant.'
        super().__init__(error)


class NoSuchBase(Exception):
    def __init__(self, error=None):
        if error is None:
            error = 'No matching base.'
        super().__init__(error)


class FailedToDownloadResource(Exception):
    def __init__(self, remote_resource: str):
        error = 'Failed to download resource {}.'
        super().__init__(error)





import atexit

import injector

from labbie.ui import keys
from labbie.ui.app import presenter as app
from labbie.ui.system_tray import view


class SystemTrayIconPresenter:
    @injector.inject
    def __init__(self, view: view.SystemTrayIcon, app_presenter: app.AppPresenter):
        self._view = view
        self._app_presenter = app_presenter

        self._view.set_double_click_handler(self.on_search_triggered)
        self._view.set_search_triggered_handler(self.on_search_triggered)
        self._view.set_settings_triggered_handler(self.on_settings_triggered)
        self._view.set_about_triggered_handler(self.on_about_triggered)

        atexit.register(self._at_exit)

    def show(self):
        self._view.show()

    def _at_exit(self):
        try:
            self._view.hide()
        except RuntimeError:
            pass

    def on_search_triggered(self):
        self._app_presenter.show(keys.SearchWindowKey())

    def on_settings_triggered(self):
        self._app_presenter.show(keys.SettingsWindowKey())

    def on_about_triggered(self):
        self._app_presenter.show(keys.AboutWindowKey())



import itertools
from typing import List, Union

import injector
import loguru

from labbie import bases
from labbie import constants
from labbie import errors
from labbie import mods
from labbie import state
from labbie import result as search_result
from labbie.ui.app import presenter as app
from labbie.ui.search.widget import view
from labbie.ui.result.widget import presenter as result

logger = loguru.logger
_POSITION_FILE = 'position.txt'


class SearchPresenter:

    @injector.inject
    def __init__(
        self,
        constants_: constants.Constants,
        app_state: state.AppState,
        app_presenter: app.AppPresenter,
        bases_: bases.Bases,
        mods_: mods.Mods,
        view: view.SearchWidget,
        result_builder: injector.AssistedBuilder[result.ResultWidgetPresenter]
    ):
        self._constants = constants_
        self._app_state = app_state
        self._app_presenter = app_presenter
        self._bases = bases_
        self._mods = mods_
        self._view = view
        self._result_builder = result_builder

        self._search_id_iter = iter(itertools.count())

        self._view.set_search_mod_handler(self.on_search_mod)
        self._view.set_search_base_handler(self.on_search_base)
        self._view.set_all_handler(self.on_all)
        self._view.set_screen_capture_handler(self.on_screen_capture)

        position = None
        if (position_path := self._constants.data_dir / _POSITION_FILE).is_file():
            with position_path.open('r', encoding='utf8') as f:
                position = [int(val) for val in f.read().split()]
        self._view.set_position(position)
        self._view.set_position_path(position_path)

        # TODO(bnorick): attach to enchants so that we can update the dropdowns when enchants change
        influences = ['Shaper', 'Elder', 'Crusader', 'Redeemer', 'Hunter', 'Warlord']
        self._view.set_influence_options(influences, influences)
        self._view.set_mods(self._mods.helm_display_mods, None)
        self._view.set_bases(self._bases.helm_display_texts)

        if constants_.debug:
            self._view.set_selected_mod('Tornado Shot fires an additional secondary Projectile')

    @property
    def widget(self):
        return self._view

    def reset_position(self):
        self._view.set_position(None)

    def cleanup(self):
        pass

    def populate_view(self, results: Union[None, search_result.Result, List[search_result.Result]],
                      clear=False, switch=False):
        logger.debug(f'{results=}')
        if not results:
            return

        if clear:
            self._view.clear_results()

        if isinstance(results, search_result.Result):
            results = [results]

        for result in results:
            result_presenter = self._result_builder.build()
            result_presenter.populate_view(result)
            tab_title = result.title[:30] + ('...' if len(result.title) > 30 else '')
            self._view.add_result_tab(tab_title, result_presenter.widget, switch=switch)

    def on_search_mod(self, checked):
        try:
            self._app_state.ensure_scrape_enabled()
        except RuntimeError as e:
            self._app_presenter.show(keys.ErrorWindowKey(str(e)))
            return

        mod = self._view.mod

        league_matches = None
        if self._app_state.league_enchants.enabled:
            try:
                league_matches = self._app_state.league_enchants.find_matching_enchants(mod)
            except errors.EnchantsNotLoaded as e:
                self._app_presenter.show(keys.ErrorWindowKey(e))
                return

        daily_matches = None
        try:
            daily_matches = self._app_state.daily_enchants.find_matching_enchants(mod)
        except errors.EnchantsNotLoaded as e:
            self._app_presenter.show(keys.ErrorWindowKey(e))
            return

        result = search_result.Result(
            title=mod,
            search=mod,
            base=False,
            league_result=league_matches,
            daily_result=daily_matches
        )
        self.populate_view(result, switch=True)

    def on_search_base(self, checked):
        try:
            self._app_state.ensure_scrape_enabled()
        except RuntimeError as e:
            self._app_presenter.show(keys.ErrorWindowKey(str(e)))
            return

        base_name = self._view.base
        ilvl = int(self._view.ilvl or 0)
        influences = self._view.influences

        league_matches = None
        if self._app_state.league_enchants.enabled:
            try:
                league_matches = self._app_state.league_enchants.find_matching_bases(base_name, ilvl, influences)
            except errors.EnchantsNotLoaded as e:
                self._app_presenter.show(keys.ErrorWindowKey(e))
                return
            except errors.NoSuchBase:
                league_matches = []

        daily_matches = None
        if self._app_state.daily_enchants.enabled:
            try:
                daily_matches = self._app_state.daily_enchants.find_matching_bases(base_name, ilvl, influences)
            except errors.EnchantsNotLoaded as e:
                self._app_presenter.show(keys.ErrorWindowKey(e))
                return
            except errors.NoSuchBase:
                daily_matches = []

        ilvl = f'i{ilvl}+ ' if ilvl != 0 else ''
        influence = f'{", ".join(influences)} ' if influences else ''
        search = f'{ilvl}{influence}{base_name}'
        result = search_result.Result(
            title=search,
            search=search,
            base=True,
            league_result=league_matches,
            daily_result=daily_matches
        )
        self.populate_view(result, switch=True)

    def on_all(self, checked):
        league_enchants = self._app_state.league_enchants.enchants
        daily_enchants = self._app_state.daily_enchants.enchants

        bases_result = search_result.Result(
            title='Bases',
            search='All Bases',
            base=False,
            league_result=league_enchants,
            daily_result=daily_enchants
        )
        self.populate_view(bases_result)

        enchants_result = search_result.Result(
            title='Enchants',
            search='All Enchants',
            base=True,
            league_result=league_enchants,
            daily_result=daily_enchants
        )
        self.populate_view(enchants_result)

    def on_screen_capture(self, checked):
        self._app_presenter.screen_capture()

    def show(self):
        self._view.show()


from labbie.ui import keys  # noqa: E402



import inspect

import loguru
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import qasync

from labbie.vendor.qtmodern import windows

logger = loguru.logger
Qt = QtCore.Qt


class BaseWindow(windows.ModernWindow):
    def __init__(self,
                 widget: QtWidgets.QWidget,
                 window_flags=(Qt.Window
                               | Qt.FramelessWindowHint
                               | Qt.WindowSystemMenuHint),
                 **kwargs
                 ):
        super().__init__(w=widget, **kwargs)
        self.setWindowFlags(window_flags)

        self.widget = widget
        widget.windowTitleChanged.connect(self._widget_window_title_changed)
        if signal_close := getattr(widget, 'signal_close', None):
            self.signal_close = signal_close

    def _widget_window_title_changed(self, title):
        self.setWindowTitle(title)

    def set_buttons(self, minimize=None, maximize=None, close=None):
        kwargs = {}
        if minimize is not None:
            kwargs['minimize'] = minimize
        if maximize is not None:
            kwargs['maximize'] = maximize
        if close is not None:
            kwargs['close'] = close
        self._set_buttons(**kwargs)

    def _set_buttons(self, **kwargs):
        flags = self.windowFlags()

        hints = {
            'minimize': Qt.WindowMinimizeButtonHint,
            'maximize': Qt.WindowMaximizeButtonHint,
            'close': Qt.WindowCloseButtonHint
        }

        for name, value in kwargs.items():
            if hint := hints.get(name):
                if value:
                    flags |= hint
                else:
                    flags &= ~hint
            else:
                raise ValueError(f'Invalid keyword argument "{name}"')

        self.setWindowFlags(flags)

    def show(self):
        super().show()
        self.setWindowState(self.windowState() & ~Qt.WindowMinimized)
        self.raise_()
        self.activateWindow()

    def toggle(self):
        if not self.isVisible():
            self.show()
        else:
            self.hide()

    @staticmethod
    def _connect_signal_to_slot(signal, slot):
        if inspect.iscoroutinefunction(slot):
            slot = qasync.asyncSlot()(slot)
        signal.connect(slot)


class BaseWidget(QtWidgets.QWidget):
    signal_close = QtCore.pyqtSignal(name='close')

    def center_on_screen(self, adjust_size=True):
        if adjust_size:
            self.adjustSize()

        # find the topmost window and move it
        last = self
        parent = self.parent()
        while parent:
            last = parent
            parent = parent.parent()

        last.move(QtWidgets.QApplication.instance().desktop().screen().rect().center() - last.rect().center())

    @staticmethod
    def _connect_signal_to_slot(signal, slot):
        if inspect.iscoroutinefunction(slot):
            slot = qasync.asyncSlot()(slot)
        signal.connect(slot)

    def closeEvent(self, event: QtGui.QCloseEvent):
        logger.debug('closeEvent, signalling')
        self.signal_close.emit()
        event.accept()



import injector

from labbie.ui.error.widget import presenter as widget
from labbie.ui.error.window import view


class ErrorWindowPresenter:

    @injector.inject
    def __init__(self, widget_presenter: widget.ErrorPresenter,
                 view_builder: injector.AssistedBuilder[view.ErrorWindow]):
        self._widget_presenter = widget_presenter
        self._view = view_builder.build(widget=widget_presenter.widget)

        self._view.signal_close.connect(self.close)

    def populate_view(self, exception):
        self._widget_presenter.populate_view(exception)

    def show(self):
        self._view.show()

    def close(self):
        self._widget_presenter.cleanup()
        self._view.hide()
        self._view.close()

    def add_close_callback(self, callback):
        self._view.signal_close.connect(callback)



import injector
import loguru

from labbie.ui.error.widget import view

logger = loguru.logger


class ErrorPresenter:

    @injector.inject
    def __init__(self, view: view.ErrorWidget):
        self._view = view

    @property
    def widget(self):
        return self._view

    def cleanup(self):
        pass

    def populate_view(self, exception):
        logger.exception(exception)
        self._view.set_error(str(exception))

    def show(self):
        self._view.show()



import asyncio
import collections
import dataclasses
import datetime
import enum
import gzip
import io
import pathlib
import re
from typing import List, NamedTuple, Optional

import aiohttp
import loguru
import orjson

from labbie import constants
from labbie import errors
from labbie import mixins

logger = loguru.logger
_FILENAME_FORMAT = '{date:%Y-%m-%d}.json.gz'
_URL_FORMAT = f'https://labbie.blob.core.windows.net/enchants/{{type}}/{_FILENAME_FORMAT}'
_VALUE_PATTERN = re.compile(r'-?\d+')
_HISTORICAL_DAYS = 15
_REFRESH_DELAY = 5 * 60  # 5 minutes
_Constants = constants.Constants


class Helm(NamedTuple):

    item_name: str
    item_base: Optional[str]
    ilvl: Optional[int]
    influences: List[str]
    unique: bool


@dataclasses.dataclass
class Enchant:

    account: str
    character: str
    item_name: str
    item_base: str
    display_name: str = dataclasses.field(init=False, default=None)
    ilvl: int
    influences: List[str]
    unique: bool
    mods: List[str]

    def __post_init__(self):
        self.display_name = self.item_name if self.unique else self.item_base

    def matches_helm(self, target: Helm):
        if target.unique:
            if self.item_name != target.item_name:
                return False
        else:
            if self.item_base != target.item_base:
                return False

        if self.ilvl < target.ilvl:
            return False

        for influence in target.influences:
            if influence not in self.influences:
                return False

        return True


class State(enum.Enum):

    DISABLED = 'Disabled via settings'
    DOWNLOADING = 'Downloading latest'
    LOADING = 'Loading'
    LOADED = 'Loaded'
    MISSING = 'Missing enchant data'


@dataclasses.dataclass
class Enchants(mixins.ObservableMixin):

    type: str
    state: State = State.DISABLED
    enchants: Optional[List[Enchant]] = None
    date: Optional[datetime.date] = None

    def __post_init__(self):
        super().__init__()
        self._refresh_task = None

    def set_enchants(self, date: Optional[datetime.date], enchants: Optional[List[Enchant]]):
        if enchants is not None:
            self.state = State.LOADED
        else:
            self.state = State.DISABLED
            if self._refresh_task:
                self._refresh_task.cancel()
                self._refresh_task = None
        self.enchants = enchants
        self.date = date
        self.notify(enchants=enchants, date=date, _log=False)

    def refresh_needed(self):
        return self.date != datetime.date.today()

    async def download_or_load(self, constants: _Constants):
        logger.info(f'starting refresh task for {self.type}')
        self._refresh_task = asyncio.create_task(self._download_when_available(constants))

        cache_dir = constants.helm_enchants_dir / self.type
        today = today_utc()

        # first try to load from cache if most recent data is cached
        try:
            self.state = State.LOADING
            date, enchants = load_enchants(cache_dir, date=today)
            self.set_enchants(date, enchants)
            return
        except errors.EnchantDataNotFound:
            pass

        # either most recent data cache was corrupt or we don't have it, load most recent cached if it's
        # the most recent downloadable, otherwise download latest
        first_cached, last_cached = cached_dates(cache_dir)
        last_downloadable = await last_downloadable_date(constants.user_agent, self.type, past_days=_HISTORICAL_DAYS)

        # try to load (fresh) cached content if we can
        if last_cached is not None:
            date = last_cached
            earliest_date = first_cached if last_downloadable is None else last_downloadable
            while date >= earliest_date:
                try:
                    self.state = State.LOADING
                    date, enchants = load_enchants(cache_dir, date=date)
                    self.set_enchants(date, enchants)
                    return
                except errors.EnchantDataInvalid:
                    date -= datetime.timedelta(days=1)
                except errors.EnchantDataNotFound:
                    # ignore missing files
                    break

        if last_downloadable is None:
            self.state = State.MISSING
            return

        try:
            self.state = State.DOWNLOADING
            date, enchants = await download(cache_dir, self.type, constants.user_agent,
                                            past_days=_HISTORICAL_DAYS)
            self.set_enchants(date, enchants)
        except errors.EnchantDataNotFound:
            # this shouldn't happen because we checked the last downloadable date and one existed within
            # _HISTORICAL_DAYS
            logger.debug(f'{first_cached=} {last_cached=} {last_downloadable=}')
            raise

    async def _download_when_available(self, constants: _Constants):
        await asyncio.sleep(15)
        cache_dir = constants.helm_enchants_dir / self.type
        while True:
            if self.date == today_utc():
                now = datetime.datetime.now(datetime.timezone.utc)
                tomorrow = (now + datetime.timedelta(days=1)).replace(hour=0, minute=0, second=0)
                duration = (tomorrow - now).total_seconds()
                logger.info(f'{self.type} scrape is fresh, sleeping until tomorrow ({duration=})')
                await asyncio.sleep(duration)

            while True:
                logger.info(f'checking if fresh {self.type} scrape is downloadable')
                available = await last_downloadable_date(constants.user_agent, self.type, 1)
                if available:
                    self.state = State.DOWNLOADING
                    date, enchants = await download(cache_dir, self.type, constants.user_agent,
                                                    past_days=_HISTORICAL_DAYS)
                    self.set_enchants(date, enchants)
                    break
                else:
                    await asyncio.sleep(_REFRESH_DELAY)

    @property
    def enabled(self):
        return self.state is not State.DISABLED

    @property
    def bases(self):
        if self.state is not State.LOADED:
            raise errors.EnchantsNotLoaded

        return set(enchant.display_name for enchant in self.enchants)

    @property
    def mods(self):
        if self.state is not State.LOADED:
            raise errors.EnchantsNotLoaded

        mods = set()
        for enchant in self.enchants:
            mods.update(set(enchant.mods))

        return mods

    def find_matching_enchants(self, target: str):
        if self.state is not State.LOADED:
            raise errors.EnchantsNotLoaded
        return find_matching_enchants(self.enchants, target)

    def find_matching_helms(self, target: Helm):
        if self.state is not State.LOADED:
            raise errors.EnchantsNotLoaded
        return find_matching_helms(self.enchants, target)

    def find_matching_bases(self, base_name: str, ilvl: int, influences: List[str]):
        logger.debug(f'{influences=}')
        if self.state is not State.LOADED:
            raise errors.EnchantsNotLoaded

        for enchant in self.enchants:
            if enchant.display_name == base_name:
                unique = enchant.unique
                item_name = enchant.item_name
                item_base = enchant.item_base
                break
        else:
            raise errors.NoSuchBase
        helm = Helm(item_name=item_name, item_base=item_base, ilvl=ilvl, influences=influences, unique=unique)
        return find_matching_helms(self.enchants, helm)


def today_utc():
    return datetime.datetime.now(datetime.timezone.utc).date()


def refresh_needed(cache_dir: pathlib.Path):
    path = cache_dir / _FILENAME_FORMAT.format(date=today_utc())
    return not path.exists()


def cached_dates(cache_dir: pathlib.Path):
    paths = sorted(cache_dir.iterdir())

    earliest_date = None
    for path in paths:
        try:
            earliest_date = datetime.date.fromisoformat(path.stem)
            break
        except ValueError:
            pass
    else:
        # short circuit if we found no valid date filenames
        return None, None

    latest_date = None
    for path in reversed(paths):
        try:
            latest_date = datetime.date.fromisoformat(path.stem)
            break
        except ValueError:
            pass

    return earliest_date, latest_date


async def last_downloadable_date(user_agent: str, type_: str, past_days):
    today = today_utc()
    async with aiohttp.ClientSession(headers={'User-Agent': user_agent}) as session:
        for days_back in range(past_days):
            date = today - datetime.timedelta(days=days_back)
            async with session.head(_URL_FORMAT.format(type=type_, date=date)) as resp:
                if resp.status == 404:
                    continue

                return date
    return None


async def download(cache_dir: pathlib.Path, type_: str, user_agent: str, past_days: int):
    logger.info(f'downloading {type_} enchants')
    today = today_utc()
    async with aiohttp.ClientSession(headers={'User-Agent': user_agent}) as session:
        for days_back in range(past_days):
            date = today - datetime.timedelta(days=days_back)
            async with session.get(_URL_FORMAT.format(type=type_, date=date)) as resp:
                if resp.status == 404:
                    continue

                logger.info(f'found data for {date=:%Y-%m-%d}')
                content = await resp.content.read()
                path = cache_dir / _FILENAME_FORMAT.format(date=date)
                with path.open('wb') as f:
                    f.write(content)
                with gzip.open(io.BytesIO(content)) as f:
                    decompressed_content = f.read()
                try:
                    enchants = [Enchant(*vals) for vals in orjson.loads(decompressed_content)]
                except orjson.JSONDecodeError:
                    logger.error(f'Invalid enchant data downloaded for {date=}')
                    path.unlink()
                    continue
                return date, enchants

        logger.error(f'no data found for the last {past_days} days, aborting')
        # if nothing was found for the window, raise
        raise errors.EnchantDataNotFound


def load_enchants(cache_dir: pathlib.Path, date=None):
    if date is not None:
        dates = (date, )
    else:
        today = today_utc()
        dates = (today - datetime.timedelta(days=days_back) for days_back in range(15))

    for date in dates:
        path = cache_dir / _FILENAME_FORMAT.format(date=date)
        if path.exists():
            try:
                with gzip.open(path) as f:
                    content = f.read().decode('utf8')
                return date, [Enchant(*vals) for vals in orjson.loads(content)]
            except orjson.JSONDecodeError:
                # delete broken files when we find them
                path.unlink()

    raise errors.EnchantDataNotFound


def find_matching_enchants(enchants: List[Enchant], target: str):
    target = target.lower()
    logger.info(f'{target=}')
    matches = []
    for enchant in enchants:
        for mod in enchant.mods:
            if target in mod.lower() or target in inexact_mod(mod).lower():
                matches.append(enchant)
    return matches


def find_matching_helms(enchants: List[Enchant], target: Helm):
    matches = []
    for enchant in enchants:
        if enchant.matches_helm(target):
            matches.append(enchant)
    return matches


def inexact_mod(mod):
    return _VALUE_PATTERN.sub('#', mod)


def enchant_summary(enchants: List[Enchant]):
    items = collections.Counter()
    influences = collections.defaultdict(collections.Counter)
    ilvls = collections.defaultdict(list)
    rare_bases = []

    for enchant in enchants:
        items[enchant.display_name] += 1
        if not enchant.unique:
            influences[enchant.item_base][', '.join(enchant.influences) or 'Uninfluenced'] += 1
            ilvls[enchant.item_base].append(enchant.ilvl)
            rare_bases.append(enchant.item_base)

    summary = []

    summary.append('Bases')
    summary.extend(f'  {count:>3d} {val}' for val, count in items.most_common())

    if rare_bases:
        rare_bases = [val for val, _ in items.most_common() if val in rare_bases]

        summary.append('')
        for base in rare_bases:
            summary.append(f'{base} ({items[base]})')
            summary.append('  Influence')
            summary.extend(f'    {count:>3d} {val}' for val, count in influences[base].most_common())

            summary.append('')
            summary.append('  Item Level')
            item_levels = sorted(ilvls[base], reverse=True)
            last = item_levels[0]
            appended = False
            count = 0
            for count, ilvl in enumerate(item_levels[1:], start=1):
                if ilvl == last:
                    continue

                prefix = '>=' if appended else '  '
                summary.append(f'    {count:>3d} {prefix}{last:>3d}')
                appended = True
                last = ilvl

            count += 1
            prefix = '>=' if appended else '  '
            summary.append(f'    {count:>3d} {prefix}{last:>3d}')
            summary.append('')
    return '\n'.join(summary)


def base_summary(enchants: List[Enchant]):
    mods = collections.Counter()

    for enchant in enchants:
        for mod in enchant.mods:
            mods[inexact_mod(mod)] += 1

    summary = [f'  {count:>3d} {val}' for val, count in mods.most_common()]

    return '\n'.join(summary)



from PyQt5 import QtCore
from PyQt5 import QtWidgets

from labbie.ui import base
from labbie.ui import switch
from labbie.ui import utils

Qt = QtCore.Qt


class SettingsWidget(base.BaseWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        group_builds = QtWidgets.QGroupBox('Извличане', self)

        lbl_league = QtWidgets.QLabel('Лига', self)
        self.switch_league = switch.Switch(self, thumb_radius=8, track_radius=5)

        lbl_daily = QtWidgets.QLabel('Дневни', self)
        self.switch_daily = switch.Switch(self, thumb_radius=8, track_radius=5)

        group_screen_capture = QtWidgets.QGroupBox('Заснемане на екрана', self)

        lbl_hotkey = QtWidgets.QLabel('Hotkey', self)
        self.edit_hotkey = QtWidgets.QLineEdit(self)

        group_bounds = QtWidgets.QGroupBox('Граници на заснемане', self)
        lbl_bounds_left = QtWidgets.QLabel('Ляво част', self)
        self.edit_left = QtWidgets.QLineEdit(self)
        lbl_bounds_top = QtWidgets.QLabel('Горна част', self)
        self.edit_top = QtWidgets.QLineEdit(self)
        lbl_bounds_right = QtWidgets.QLabel('Дясна част', self)
        self.edit_right = QtWidgets.QLineEdit(self)
        lbl_bounds_bottom = QtWidgets.QLabel('Долна част', self)
        self.edit_bottom = QtWidgets.QLineEdit(self)
        self.btn_select_bounds = QtWidgets.QPushButton('Избери', self)

        self.btn_reset_window_positions = QtWidgets.QPushButton('Нулиране на позициите на прозореца', self)
        size = self.btn_reset_window_positions.sizeHint()
        self.btn_reset_window_positions.setMinimumWidth(size.width() + 12)

        self.btn_save = QtWidgets.QPushButton('Запази', self)
        self.btn_cancel = QtWidgets.QPushButton('Отказ', self)

        layout_active = QtWidgets.QGridLayout()
        layout_active.addWidget(lbl_league, 0, 0)
        layout_active.addWidget(self.switch_league, 0, 1)
        layout_active.addWidget(lbl_daily, 1, 0)
        layout_active.addWidget(self.switch_daily, 1, 1)

        layout_builds = QtWidgets.QHBoxLayout()
        layout_builds.addLayout(layout_active)
        layout_builds.addStretch(1)
        group_builds.setLayout(layout_builds)

        lbl_clear = QtWidgets.QLabel('Изчистване на предишни резултати', self)
        self.switch_clear = switch.Switch(self, thumb_radius=8, track_radius=5)

        layout_capture_general = QtWidgets.QGridLayout()
        layout_capture_general.addWidget(lbl_hotkey, 0, 0)
        layout_capture_general.addWidget(self.edit_hotkey, 0, 1)
        layout_capture_general.addWidget(lbl_clear, 1, 0)
        layout_capture_general.addWidget(self.switch_clear, 1, 1)

        layout_capture_top = QtWidgets.QHBoxLayout()
        layout_capture_top.addLayout(layout_capture_general)
        layout_capture_top.addStretch(1)

        layout_bounds_vals_grid = QtWidgets.QGridLayout()
        layout_bounds_vals_grid.addWidget(lbl_bounds_left, 0, 0, Qt.AlignmentFlag.AlignRight)
        layout_bounds_vals_grid.addWidget(self.edit_left, 0, 1)
        layout_bounds_vals_grid.addWidget(lbl_bounds_top, 0, 2, Qt.AlignmentFlag.AlignRight)
        layout_bounds_vals_grid.addWidget(self.edit_top, 0, 3)
        layout_bounds_vals_grid.addWidget(lbl_bounds_right, 1, 0, Qt.AlignmentFlag.AlignRight)
        layout_bounds_vals_grid.addWidget(self.edit_right, 1, 1)
        layout_bounds_vals_grid.addWidget(lbl_bounds_bottom, 1, 2, Qt.AlignmentFlag.AlignRight)
        layout_bounds_vals_grid.addWidget(self.edit_bottom, 1, 3)

        layout_bounds_vals = QtWidgets.QHBoxLayout()
        layout_bounds_vals.addLayout(layout_bounds_vals_grid)
        layout_bounds_vals.addStretch(1)

        layout_select = QtWidgets.QHBoxLayout()
        layout_select.addWidget(self.btn_select_bounds)
        layout_select.addStretch(1)

        layout_bounds = QtWidgets.QVBoxLayout()
        layout_bounds.addLayout(layout_bounds_vals)
        layout_bounds.addLayout(layout_select)

        group_bounds.setLayout(layout_bounds)

        layout_screen_capture = QtWidgets.QVBoxLayout()
        layout_screen_capture.addLayout(layout_capture_top)
        layout_screen_capture.addWidget(group_bounds)

        group_screen_capture.setLayout(layout_screen_capture)

        layout_actions = QtWidgets.QHBoxLayout()
        layout_actions.addWidget(self.btn_reset_window_positions)
        layout_actions.addStretch(1)
        layout_actions.addWidget(self.btn_save)
        layout_actions.addWidget(self.btn_cancel)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(group_builds)
        layout.addWidget(group_screen_capture)
        layout.addLayout(layout_actions)
        self.setLayout(layout)

        self.btn_cancel.clicked.connect(self.close)

        self.setWindowTitle('Настройки')
        self.center_on_screen()

    def set_select_button_text(self, text):
        self.btn_select_bounds.setText(text)

    def set_select_bounds_handler(self, handler):
        self._connect_signal_to_slot(self.btn_select_bounds.clicked, handler)

    def set_reset_window_positions_handler(self, handler):
        self._connect_signal_to_slot(self.btn_reset_window_positions.clicked, handler)

    def set_save_handler(self, handler):
        self._connect_signal_to_slot(self.btn_save.clicked, handler)

    @utils.checkbox_property
    def league(self):
        return self.switch_league

    @utils.checkbox_property
    def daily(self):
        return self.switch_daily

    @utils.checkbox_property
    def clear_previous(self):
        return self.switch_clear

    @utils.text_property
    def hotkey(self):
        return self.edit_hotkey

    @utils.text_property
    def left(self):
        return self.edit_left

    @utils.text_property
    def top(self):
        return self.edit_top

    @utils.text_property
    def right(self):
        return self.edit_right

    @utils.text_property
    def bottom(self):
        return self.edit_bottom



import inspect
import os

import injector
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import qasync

Qt = QtCore.Qt
_SIZEGRIP_STYLE = '''
QSizeGrip {
    background-color: rgb(255, 255, 255);
    width: 16px;
    height: 16px;
}
'''


class ScreenSelectionWidget(QtWidgets.QWidget):

    rightClicked = QtCore.pyqtSignal()

    @injector.inject
    def __init__(self, left, top, right, bottom, **kwargs):
        super().__init__(**kwargs)
        self.__press_pos = None

        window_flags = Qt.FramelessWindowHint
        if os.name == 'posix':
            window_flags = window_flags | QtCore.Qt.Tool
        self.setWindowFlags(window_flags)
        self.setAttribute(Qt.WA_TranslucentBackground)

        # self.btn_done = QtWidgets.QPushButton('Done', self)
        lbl_done = QtWidgets.QLabel('плъзнете, за да преместите\nдесен клик запазвва', self)
        lbl_done.setFont(QtGui.QFont("Arial", 20, QtGui.QFont.Bold))
        lbl_done.setStyleSheet('QLabel { color: rgb(37, 37, 37);}')
        lbl_done.setAlignment(Qt.AlignmentFlag.AlignHCenter)


        def make_sizegrip():
            sizegrip = QtWidgets.QSizeGrip(self)
            sizegrip.setStyleSheet(_SIZEGRIP_STYLE)
            sizegrip.setMinimumSize(16, 16)
            return sizegrip

        sizegrips = [make_sizegrip() for _ in range(4)]

        top_handles = QtWidgets.QHBoxLayout()
        top_handles.setContentsMargins(0, 0, 0, 0)
        top_handles.addWidget(sizegrips[0])
        top_handles.addStretch(1)
        top_handles.addWidget(sizegrips[1])

        bottom_handles = QtWidgets.QHBoxLayout()
        bottom_handles.setContentsMargins(0, 0, 0, 0)
        bottom_handles.addWidget(sizegrips[2])
        bottom_handles.addStretch(1)
        bottom_handles.addWidget(sizegrips[3])

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(top_handles)
        layout.addStretch(1)
        layout.addWidget(lbl_done, 0, Qt.AlignmentFlag.AlignHCenter)
        layout.addStretch(1)
        layout.addLayout(bottom_handles)
        self.setLayout(layout)

        self.set_position(left, top, right, bottom)
        self.setMinimumSize(50, 50)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.__press_pos = event.pos()  # remember starting position
        elif event.button() == Qt.RightButton:
            self.rightClicked.emit()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.__press_pos = None

    def mouseMoveEvent(self, event):
        if self.__press_pos:  # follow the mouse
            self.move(self.pos() + (event.pos() - self.__press_pos))

    def paintEvent(self, event=None):
        painter = QtGui.QPainter(self)
        painter.setOpacity(0.7)
        painter.setBrush(Qt.white)
        painter.setPen(QtGui.QPen(Qt.white))
        painter.drawRect(self.rect())

    def set_position(self, left, top, right, bottom):
        self.resize(right - left, bottom - top)
        self.move(left, top)
        self._old_pos = self.pos()

    def get_position(self):
        geom = self.geometry()
        return (geom.left(), geom.top(), geom.right(), geom.bottom())

    def set_done_handler(self, handler):
        self._connect_signal_to_slot(self.rightClicked, handler)

    @staticmethod
    def _connect_signal_to_slot(signal, slot):
        if inspect.iscoroutinefunction(slot):
            slot = qasync.asyncSlot()(slot)
        signal.connect(slot)



import os

from PyQt5 import QtCore

from labbie.ui import base
from labbie.ui.about.widget import view as about_widget


class AboutWindow(base.BaseWindow):

    def __init__(self, widget: about_widget.AboutWidget):
        super().__init__(widget=widget)
        self.set_buttons(close=True)
        flags = self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint
        if os.name == 'posix':
            flags = flags | QtCore.Qt.Tool
        self.setWindowFlags(flags)



from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

Qt = QtCore.Qt


class Switch(QtWidgets.QAbstractButton):

    def __init__(self, parent=None, track_radius=10, thumb_radius=8):
        super().__init__(parent=parent)
        self.setCheckable(True)
        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._track_radius = track_radius
        self._thumb_radius = thumb_radius

        self._margin = max(0, self._thumb_radius - self._track_radius)
        self._base_offset = max(self._thumb_radius, self._track_radius)
        self._end_offset = {
            True: lambda: self.width() - self._base_offset,
            False: lambda: self._base_offset,
        }
        self._offset = self._base_offset

        palette = self.palette()
        if self._thumb_radius > self._track_radius:
            self._track_color = {
                True: palette.highlight(),
                False: palette.dark(),
            }
            self._thumb_color = {
                True: palette.highlight(),
                False: palette.light(),
            }
            self._text_color = {
                True: palette.highlightedText().color(),
                False: palette.dark().color(),
            }
            self._thumb_text = {
                True: '',
                False: '',
            }
            self._track_opacity = 0.5
        else:
            self._thumb_color = {
                True: palette.highlightedText(),
                False: palette.light(),
            }
            self._track_color = {
                True: palette.highlight(),
                False: palette.dark(),
            }
            self._text_color = {
                True: palette.highlight().color(),
                False: palette.dark().color(),
            }
            self._thumb_text = {
                True: '✔',
                False: '✕',
            }
            self._track_opacity = 1

    @QtCore.pyqtProperty(int)
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value):
        self._offset = value
        self.update()

    def sizeHint(self):  # pylint: disable=invalid-name
        return QtCore.QSize(
            4 * self._track_radius + 2 * self._margin,
            2 * self._track_radius + 2 * self._margin,
        )

    def setChecked(self, checked):
        super().setChecked(checked)
        self.offset = self._end_offset[checked]()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.offset = self._end_offset[self.isChecked()]()

    def paintEvent(self, event):  # pylint: disable=invalid-name, unused-argument
        p = QtGui.QPainter(self)
        p.setRenderHint(QtGui.QPainter.Antialiasing, True)
        p.setPen(Qt.NoPen)
        track_opacity = self._track_opacity
        thumb_opacity = 1.0
        text_opacity = 1.0
        if self.isEnabled():
            track_brush = self._track_color[self.isChecked()]
            thumb_brush = self._thumb_color[self.isChecked()]
            text_color = self._text_color[self.isChecked()]
        else:
            track_opacity *= 0.8
            track_brush = self.palette().shadow()
            thumb_brush = self.palette().mid()
            text_color = self.palette().shadow().color()

        p.setBrush(track_brush)
        p.setOpacity(track_opacity)
        p.drawRoundedRect(
            self._margin,
            self._margin,
            self.width() - 2 * self._margin,
            self.height() - 2 * self._margin,
            self._track_radius,
            self._track_radius,
        )
        p.setBrush(thumb_brush)
        p.setOpacity(thumb_opacity)
        p.drawEllipse(
            self.offset - self._thumb_radius,
            self._base_offset - self._thumb_radius,
            2 * self._thumb_radius,
            2 * self._thumb_radius,
        )
        p.setPen(text_color)
        p.setOpacity(text_opacity)
        font = p.font()
        font.setPixelSize(1.5 * self._thumb_radius)
        p.setFont(font)
        p.drawText(
            QtCore.QRectF(
                self.offset - self._thumb_radius,
                self._base_offset - self._thumb_radius,
                2 * self._thumb_radius,
                2 * self._thumb_radius,
            ),
            Qt.AlignCenter,
            self._thumb_text[self.isChecked()],
        )

    def mouseReleaseEvent(self, event):  # pylint: disable=invalid-name
        super().mouseReleaseEvent(event)
        if event.button() == Qt.LeftButton:
            anim = QtCore.QPropertyAnimation(self, b'offset', self)
            anim.setDuration(120)
            anim.setStartValue(self.offset)
            anim.setEndValue(self._end_offset[self.isChecked()]())
            anim.start()

    def enterEvent(self, event):  # pylint: disable=invalid-name
        self.setCursor(Qt.PointingHandCursor)
        super().enterEvent(event)


class Toggle(Switch):

    def __init__(self, parent=None, track_radius=10, thumb_radius=8):
        super().__init__(parent=parent, track_radius=track_radius, thumb_radius=thumb_radius)
        palette = self.palette()
        self._track_color = {
            True: palette.dark(),
            False: palette.dark(),
        }
        self._thumb_color = {
            True: palette.light(),
            False: palette.light(),
        }
        self._text_color = {
            True: palette.highlightedText().color(),
            False: palette.dark().color(),
        }
        self._thumb_text = {
            True: '',
            False: '',
        }
        self._track_opacity = 0.5


import inspect

import injector
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import qasync

Qt = QtCore.Qt


class TransparentWindow(QtWidgets.QWidget):

    @injector.inject
    def __init__(self, close_on_click: bool = False, **kwargs):
        super().__init__(**kwargs)
        self._close_on_click = close_on_click

        self.setWindowFlags(Qt.FramelessWindowHint)
        self.setAttribute(Qt.WA_TranslucentBackground)

    def mousePressEvent(self, event):
        if self._close_on_click:
            event.ignore()
            self.close()

    def paintEvent(self, event=None):
        painter = QtGui.QPainter(self)
        painter.setOpacity(0.7)
        painter.setBrush(Qt.white)
        painter.setPen(QtGui.QPen(Qt.white))
        painter.drawRect(self.rect())

    @staticmethod
    def _connect_signal_to_slot(signal, slot):
        if inspect.iscoroutinefunction(slot):
            slot = qasync.asyncSlot()(slot)
        signal.connect(slot)



import abc
import dataclasses
import pathlib
from typing import Optional

import loguru
import toml

from labbie import mixins
from labbie import bounds

logger = loguru.logger
_Bounds = bounds.Bounds


@dataclasses.dataclass
class BaseConfig(mixins.ObservableMixin, abc.ABC):
    @classmethod
    @abc.abstractmethod
    def load(cls, base_path: pathlib.Path):
        raise NotImplementedError


@dataclasses.dataclass
class HotkeysConfig(mixins.ObservableMixin, mixins.SerializableMixin):
    ocr: Optional[str] = '`'

    # TODO(bnorick): figure out how to hook notify into ocr changes best, so we don't need to
    # manually notify when we change it (or maybe that's best, who knows)

    @classmethod
    def from_dict(cls, config_dict):
        # handle 'none' in toml to => None
        for field_ in dataclasses.fields(cls):
            key = field_.name
            v = config_dict.get(key, '')
            if type(v) == str and v.lower() == 'none':
                config_dict[key] = None

        return super().from_dict(config_dict)


@dataclasses.dataclass
class OcrConfig(mixins.SerializableMixin):
    clear_previous: bool = True
    bounds: _Bounds = _Bounds(left=335, top=210, right=916, bottom=455)


@dataclasses.dataclass
class UiConfig(mixins.SerializableMixin):
    hotkeys: HotkeysConfig = HotkeysConfig()


@dataclasses.dataclass
class Config(BaseConfig, mixins.SerializableMixin):
    ui: UiConfig = UiConfig()
    ocr: OcrConfig = OcrConfig()
    league: bool = True
    daily: bool = True

    @classmethod
    def from_toml(cls, path: pathlib.Path):
        with path.open() as f:
            config_dict = toml.load(f)
        config = cls.from_dict(config_dict)
        config._path = path
        return config

    @classmethod
    def load(cls, base_path: pathlib.Path):
        path = base_path / 'config.toml'
        logger.info(f'Loading config from {path}')
        return cls.from_toml(path)

    def save(self):
        backup_path = self._path.with_suffix('.toml.bak')
        if backup_path.exists():
            backup_path.unlink()

        self._path.rename(backup_path)
        with self._path.open('w', encoding='utf8') as f:
            toml.dump(dataclasses.asdict(self), f)



import dataclasses
import enum
from typing import Optional

import injector

from labbie import enchants
from labbie import mixins


class State(enum.Enum):
    STARTING = 'Starting'
    OCR = 'Scanning enchants'
    READY = 'Ready'
    ERROR = 'Error'


@injector.singleton
@dataclasses.dataclass
class AppState(mixins.ObservableMixin):
    state: State = State.STARTING
    resources_ready: bool = False
    league_enchants: enchants.Enchants = enchants.Enchants('league')
    daily_enchants: enchants.Enchants = enchants.Enchants('daily')
    last_error: Optional[str] = None

    def ensure_scrape_enabled(self):
        league_disabled = self.league_enchants.state is enchants.State.DISABLED
        daily_disabled = self.daily_enchants.state is enchants.State.DISABLED
        if league_disabled and daily_disabled:
            raise RuntimeError('No enchant scrapes are enabled, please edit the settings.')

    def __setattr__(self, name, val):
        cls_fields = {f.name for f in dataclasses.fields(self)}
        if name not in cls_fields:
            object.__setattr__(self, name, val)
            return

        cur_val = getattr(self, name)
        if cur_val != val:
            object.__setattr__(self, name, val)
            self.notify(**{name: val})



import asyncio
import dataclasses
import functools
import gzip
import pathlib
from typing import ClassVar, Dict, List, Optional, Tuple, Union

import aiohttp
import injector
from loguru import logger
import orjson

from labbie import constants
from labbie import errors
from labbie import state
from labbie import utils

_Constants = constants.Constants
_LOADERS = {}


def loader(type_):
    def decorator(fn):
        _LOADERS[type_] = fn
        return fn
    return decorator


@loader('.json.gz')
def load_json_gz(path: pathlib.Path):
    with gzip.open(path) as f:
        content = f.read().decode('utf8')
    return orjson.loads(content)


@loader('.json')
def load_json(path: pathlib.Path):
    with path.open(encoding='utf8') as f:
        return orjson.loads(f.read())


def get_loader(path: pathlib.Path):
    suffixes = path.suffixes

    for i in range(len(suffixes)):
        type_ = ''.join(suffixes[i:])
        if loader := _LOADERS.get(type_):
            return loader

    raise ValueError(f'No loader specified for file with suffixes {suffixes}')


@dataclasses.dataclass
class Resource:
    _CONTAINER_URL: ClassVar[str] = 'https://labbie.blob.core.windows.net/resources'

    version: int
    path_format: str

    @functools.cached_property
    def path(self) -> pathlib.Path:
        return pathlib.Path(self.path_format.format(version=self.version))

    @functools.cached_property
    def url(self) -> pathlib.Path:
        return f'{self._CONTAINER_URL}/{self.path}'

    @functools.cached_property
    def name(self) -> pathlib.Path:
        return self.path.name

    def local_path(self, resources_dir: pathlib.Path):
        return resources_dir / self.name

    def cached_hash_path(self, resources_dir: pathlib.Path):
        local_path = self.local_path(resources_dir)
        return local_path.parent / f'{local_path.name}.md5'

    async def needs_update(self, resources_dir: pathlib.Path,
                           session: Optional[aiohttp.ClientSession] = None):
        async with utils.client_session(session) as session:
            async with session.head(self.url) as resp:
                if resp.status == 404:
                    return True
                cached_hash = self.cached_hash(resources_dir)
                remote_hash = resp.headers['Content-MD5']
                needs_update = cached_hash != remote_hash
                logger.debug(f'{self.local_path(resources_dir)} {needs_update=} '
                             f'{cached_hash=}{"==" if not needs_update else "!="}{remote_hash=}')
                return needs_update

    def cached_hash(self, resources_dir: pathlib.Path):
        hash_path = self.cached_hash_path(resources_dir)

        if not hash_path.exists():
            return None

        with hash_path.open(encoding='utf8') as f:
            return f.read()

    async def load_or_download(self, resources_dir: pathlib.Path, force: bool = False,
                               session: Optional[aiohttp.ClientSession] = None):
        needs_update = await self.needs_update(resources_dir, session=session)
        if force or needs_update:
            async with utils.client_session(session) as session:
                async with session.get(self.url) as resp:
                    if resp.status != 200:
                        raise errors.FailedToDownloadResource(self.url)

                    content = await resp.content.read()
                    hash = resp.headers['Content-MD5']
                    self.save(resources_dir, content, hash)

        return self.load(resources_dir)

    def save(self, resources_dir: pathlib.Path, content: Union[str, bytes], hash: str):
        kwargs = {}
        if isinstance(content, str):
            mode = 'w'
            kwargs['encoding'] = 'utf8'
        elif isinstance(content, bytes):
            mode = 'wb'
        else:
            raise ValueError(f'Invalid content type, expected str or bytes but got {type(content).__name__}')

        local_path = self.local_path(resources_dir)
        with local_path.open(mode, **kwargs) as f:
            f.write(content)

        hash_path = self.cached_hash_path(resources_dir)
        with hash_path.open('w', encoding='utf8') as f:
            f.write(hash)

    def load(self, resources_dir: pathlib.Path):
        local_path = self.local_path(resources_dir)
        loader = get_loader(local_path)
        return loader(local_path)


@injector.singleton
class ResourceManager:

    # NOTE: file names must not collide, as the local path only includes the name and not the full path
    _RESOURCES = {
        'trade_stats': Resource(version=3, path_format='pathofexile/{version}/stats.json.gz'),
        'items': Resource(version=3, path_format='pathofexile/{version}/items.json.gz'),
        'mods': Resource(version=3, path_format='repoe/{version}/mods.json.gz'),
    }

    @injector.inject
    def __init__(self, constants: _Constants, app_state: state.AppState):
        self._constants = constants
        self._app_state = app_state

        self._init_task = None

        # NOTE: the following attributes are set by _get_all_resources
        self.trade_stats: Dict[str, str] = None
        self.items: Dict[str, List[Tuple[bool, str, str]]] = None
        self.mods: Dict[str, List[List[str, List[str], List[Union[float, int, str]], bool]]] = None

    def initialize(self):
        self._constants.resources_dir.mkdir(parents=True, exist_ok=True)
        self._init_task = asyncio.create_task(self._get_all_resources())

    async def _get_all_resources(self):
        # TODO(bnorick): handle potential PermissionError / OSError from disk write failures in a reasonable way
        async with aiohttp.ClientSession() as session:
            for name, resource in self._RESOURCES.items():
                value = await resource.load_or_download(
                    resources_dir=self._constants.resources_dir, session=session)
                setattr(self, name, value)
        self._app_state.resources_ready = True



import argparse
import asyncio
import os
import sys

import injector
import loguru
from PyQt5 import QtGui
from PyQt5 import QtWidgets
import qasync

from labbie import config
from labbie import constants
from labbie import resources
from labbie import state
from labbie import utils
from labbie.di import module
from labbie.ui import utils as ui_utils
from labbie.ui.app import presenter as app
from labbie.vendor.qtmodern import styles

logger = loguru.logger
_Module = module.Module
_Injector = injector.Injector
_Config = config.Config
_Constants = constants.Constants


def parse_args():
    parser = argparse.ArgumentParser('Labbie')
    parser.add_argument('--debug', action='store_true', help='Enable debug mode')
    return parser.parse_args()


def main():
    utils.logs_dir().mkdir(exist_ok=True, parents=True)
    utils.exit_if_already_running()

    args = parse_args()
    if args.debug:
        os.environ['LABBIE_DEBUG'] = '1'

    logger.remove()
    log_path = utils.logs_dir() / 'current_run.log'
    if log_path.is_file():
        prev_log_path = log_path.with_name('prev_run.log')
        # delete prev log if it exists
        try:
            prev_log_path.unlink()
        except FileNotFoundError:
            pass
        # move log from last run to prev log
        try:
            log_path.rename(prev_log_path)
        except FileExistsError:
            pass
    log_filter = utils.LogFilter('INFO')
    logger.add(log_path, filter=log_filter, rotation="100 MB", retention=1, mode='w', encoding='utf8')

    app = QtWidgets.QApplication(sys.argv)
    styles.dark(app)
    app.setQuitOnLastWindowClosed(False)

    ui_utils.fix_taskbar_icon()
    icon_path = ui_utils.asset_path('taxi.png')
    app.setWindowIcon(QtGui.QIcon(str(icon_path)))

    loop = qasync.QSelectorEventLoop(app)
    asyncio.set_event_loop(loop)

    with loop:
        loop.run_until_complete(start(log_filter))
        sys.exit(loop.run_forever())


async def focus_if_other_instances(app_presenter):
    mm = utils.instances_shm()
    mm[0] = 1
    while True:
        if mm[0] > 1:
            logger.info(f'instances={mm[0]}')
            app_presenter.foreground()
            mm[0] = 1
        await asyncio.sleep(0.1)


async def start(log_filter):
    injector = _Injector(_Module())

    constants = injector.get(_Constants)

    if constants.debug:
        log_filter.level = 'DEBUG'

    # ensure directories exist
    constants.screenshots_dir.mkdir(parents=True, exist_ok=True)
    (constants.helm_enchants_dir / 'league').mkdir(parents=True, exist_ok=True)
    (constants.helm_enchants_dir / 'daily').mkdir(parents=True, exist_ok=True)

    config = injector.get(_Config)

    resource_manager = injector.get(resources.ResourceManager)
    resource_manager.initialize()
    await resource_manager._init_task

    app_state = injector.get(state.AppState)
    if config.league:
        asyncio.create_task(app_state.league_enchants.download_or_load(constants))
    if config.daily:
        asyncio.create_task(app_state.daily_enchants.download_or_load(constants))

    app_presenter = injector.get(app.AppPresenter)
    app_presenter.launch()
    asyncio.create_task(focus_if_other_instances(app_presenter))


if __name__ == '__main__':
    main()



import dataclasses
import functools
import os
import pathlib
from typing import ClassVar, Union

import dacite
import loguru
import toml

from labbie import mixins, version
from labbie import utils

logger = loguru.logger
_DEFAULT_DATA_DIR = utils.default_data_dir()
_DEFAULT_CONFIG_DIR = utils.default_config_dir()
_MISSING = object()


@dataclasses.dataclass(frozen=True)
class BaseConstants(mixins.SerializableMixin):
    @classmethod
    def load(cls, _constants_path: Union[None, str, pathlib.Path] = _MISSING, **kwargs):
        """Load constants.

        There are three potential sources for constant values, in order of their preference
        (from highest to lowest):
            1) Environment variables with names matching "LABBIE_[CONSTANT_ENV_VAR_NAME]", e.g.,
            LABBIE_DATA_DIR corresponds with the data_dir constant.
            2) Keyword arguments.
            3) Constants toml file.
        """
        if _constants_path is not _MISSING:
            path_source = 'kwarg'
            if _constants_path:
                _constants_path = pathlib.Path(_constants_path)
        else:
            path_source = 'default'
            _constants_path = cls._DEFAULT_PATH

        valid_names = {field.name for field in dataclasses.fields(cls)}

        explicit_vals = {}
        for name, val in kwargs.items():
            if name in valid_names:
                logger.info(f'Overriding constant "{name}" from kwarg, {val=}.')
                explicit_vals[name] = val
            else:
                raise ValueError(f'Invalid kwarg "{name}", no such field.')

        env_prefix = 'labbie_'
        if extra_prefix := getattr(cls, '_ENV_PREFIX', None):
            env_prefix += f'{extra_prefix.lower()}_'
        for name, val in os.environ.items():
            if name.lower().startswith(env_prefix):
                field_name = name.lower()[len(env_prefix):]  # remove the prefix
                if field_name in valid_names:
                    logger.info(f'Overriding constant "{cls.__name__}.{field_name}" from environment variable ({name}), {val=}.')
                    explicit_vals[field_name] = val
                elif field_name == 'constants':
                    path_source = 'env var'
                    _constants_path = pathlib.Path(val)

        if _constants_path:
            # when constants path is explicitly set, the corresponding file must exist
            if _constants_path.is_file():
                logger.info(f'Loading constants from {_constants_path} specified by {path_source} (with {len(explicit_vals)} overrides).')
                constants = cls.from_toml(_constants_path, overrides=explicit_vals)
            elif path_source != 'default':
                    raise ValueError(f'The path provided to load {cls.__name__} is not a file: {_constants_path}')
            else:
                logger.info('Loading constants from environment variables and kwargs.')
                constants = cls.from_dict(explicit_vals)
        else:
            logger.info('Loading constants from environment variables and kwargs.')
            constants = cls.from_dict(explicit_vals)

        logger.debug(f'{constants=}')

        return constants

    @classmethod
    def from_toml(cls, path: pathlib.Path, overrides=None):
        with path.open() as f:
            d = toml.load(f)
        if overrides:
            d.update(overrides)
        return cls.from_dict(d)

    @classmethod
    def from_dict(cls, d):
        config = dacite.Config(
            cast=[pathlib.Path],
            type_hooks={bool: lambda v: v if isinstance(v, bool) else v.lower() in ('t', 'true', '1')}
        )
        return super().from_dict(d, config=config)


@dataclasses.dataclass(frozen=True)
class Constants(BaseConstants):
    _DEFAULT_PATH: ClassVar[pathlib.Path] = utils.default_config_dir() / 'constants.toml'

    debug: bool = False
    data_dir: pathlib.Path = _DEFAULT_DATA_DIR
    config_dir: pathlib.Path = _DEFAULT_CONFIG_DIR
    user_agent: str = f'Labbie v{version.__version__}'
    dilate: bool = False

    @functools.cached_property
    def logs_dir(self):
        return utils.logs_dir()

    @functools.cached_property
    def helm_enchants_dir(self):
        return self.data_dir / 'helm'

    @functools.cached_property
    def screenshots_dir(self):
        return self.data_dir / 'screenshots'

    @functools.cached_property
    def resources_dir(self):
        return self.data_dir / 'resources'

    @classmethod
    def from_toml(cls, path: pathlib.Path, overrides=None):
        with path.open() as f:
            d = toml.load(f)
        if overrides:
            d.update(overrides)
        return cls.from_dict(d)



import os

from PyQt5 import QtCore

from labbie.ui import base
from labbie.ui.search.widget import view as search_widget


class SearchWindow(base.BaseWindow):

    def __init__(self, widget: search_widget.SearchWidget):
        super().__init__(widget=widget)
        self.set_buttons(minimize=False, close=True)

        flags = self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint
        if os.name == 'nt':
            flags = flags | QtCore.Qt.SubWindow
        elif os.name == 'posix':
            flags = flags | QtCore.Qt.Tool
        self.setWindowFlags(flags)

    # NOTE: If we want to add minimize button back, this makes it act like close
    # ref: https://stackoverflow.com/questions/16036336/how-to-change-minimize-event-behavior-in-pyqt-or-pyside

    # def changeEvent(self, event):
    #     if event.type() == QtCore.QEvent.WindowStateChange:
    #         if self.windowState() & QtCore.Qt.WindowMinimized:
    #             event.ignore()
    #             self.close()
    #             return

    #     super().changeEvent(event)



from PyQt5 import QtCore

from labbie.ui import base
from labbie.ui.error.widget import view as error_widget


class ErrorWindow(base.BaseWindow):

    def __init__(self, widget: error_widget.ErrorWidget):
        super().__init__(widget=widget)
        self.set_buttons(close=True)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)



from typing import Tuple, Union
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from labbie.ui import base
from labbie.ui import switch
from labbie.ui import utils
from labbie import version

Qt = QtCore.Qt
_SCRAPES_FORMAT = '''
<p style="font-weight: bold; font-size: 10pt;">Извлечени данни</p>
<table style="align: center;">
<tr>
<td width="25%"></td>
<td width="25%">
<span style="font-weight: bold;">Лига</span> {league:%Y-%m-%d}
</td>
<td width="25%" style="padding-left: 10px;">
<span style="font-weight: bold; padding-left: 10px;">Дневно</span> {daily:%Y-%m-%d}
</td>
<td width="25%"></td>
</tr>
</table>
'''

class AboutWidget(base.BaseWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        lbl_title = QtWidgets.QLabel('POE-LabHelper', self)
        lbl_title.setStyleSheet('QLabel{font-size: 12pt; font-weight: bold;}')
        lbl_version = QtWidgets.QLabel(f'v{version.__version__}', self)
        lbl_version.setStyleSheet('QLabel{font-size: 9pt; font-weight: bold;}')

        details = ('Този софтуер е достъпен само за членовете на клана POE България.<br />'
                   'За да поддържам този проект жив, трябва да върша шибана работа,<br />'
                   'така че, моля, дарете някое левче на клетия разработчик. '
                   '<a href="https://www.paypal.com/paypalme/rebranduk">Paypal</a> '
                   )
        lbl_details = QtWidgets.QLabel(details, self)
        lbl_details.setStyleSheet('QLabel{font-size: 9pt;}')
        lbl_details.setOpenExternalLinks(True)

        self.lbl_scrapes = QtWidgets.QLabel(self)
        self.lbl_scrapes.setAlignment(Qt.AlignCenter)

        self.btn_relaunch = QtWidgets.QPushButton(self)
        self.btn_relaunch.setIconSize(QtCore.QSize(24, 16))
        self.btn_open_data = QtWidgets.QPushButton(
            utils.recolored_icon('folder_space.svg', 180), 'Отвори Данни', self)
        self.btn_open_data.setIconSize(QtCore.QSize(24, 16))
        self.btn_open_logs = QtWidgets.QPushButton(
            utils.recolored_icon('folder_space.svg', 180), 'Отвори Регистър', self)
        self.btn_open_logs.setIconSize(QtCore.QSize(24, 16))

        layout_buttons = QtWidgets.QHBoxLayout()
        layout_buttons.addWidget(self.btn_relaunch)
        layout_buttons.addWidget(self.btn_open_data)
        layout_buttons.addWidget(self.btn_open_logs)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(10, 0, 10, 10)
        layout.addWidget(lbl_title, 0, Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(lbl_version, 0, Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(lbl_details)
        layout.addSpacing(5)
        layout.addWidget(self.lbl_scrapes, Qt.AlignmentFlag.AlignHCenter)
        layout.addSpacing(5)
        layout.addLayout(layout_buttons)
        self.setLayout(layout)

        self.center_on_screen()

    def update_relaunch_button(self, to_debug):
        if to_debug:
            self.btn_relaunch.setIcon(utils.recolored_icon('bug_space.svg', 180))
            self.btn_relaunch.setText('Повторно стартиране (Debug)')
        else:
            self.btn_relaunch.setIcon(utils.recolored_icon('return_space.svg', 180))
            self.btn_relaunch.setText('Повторно стартиране (Normal)')

    def set_scrapes(self, league, daily):
        self.lbl_scrapes.setText(_SCRAPES_FORMAT.format(league=league, daily=daily))
        self.lbl_scrapes.adjustSize()

    def set_relaunch_handler(self, handler):
        self._connect_signal_to_slot(self.btn_relaunch.clicked, handler)

    def set_open_data_handler(self, handler):
        self._connect_signal_to_slot(self.btn_open_data.clicked, handler)

    def set_open_logs_handler(self, handler):
        self._connect_signal_to_slot(self.btn_open_logs.clicked, handler)

    def exit(self):
        QtWidgets.QApplication.exit()



import atexit
import contextlib
import dataclasses
import mmap
import os
import pathlib
import sys
from typing import Optional
import aiohttp

import loguru

logger = loguru.logger


def root_dir():
    if getattr(sys, 'frozen', False):
        # The application is frozen
        return pathlib.Path(sys.executable).parent
    else:
        # The application is not frozen
        return pathlib.Path(__file__).parent.parent.parent


def logs_dir():
    return root_dir() / 'logs'


def assets_dir():
    return root_dir() / 'assets'


def bin_dir():
    return root_dir() / 'bin'


def default_config_dir():
    return root_dir() / 'config'


def default_data_dir():
    return root_dir() / 'data'


def relaunch(debug, exit_fn=None):
    cmd = []
    if sys.argv[0] == 'labbie':
        from labbie import __main__ as main
        cmd.append(main.__file__)
    else:
        cmd.append(f'"{sys.argv[0]}"')
    if debug:
        cmd.append('--debug')
    if not getattr(sys, 'frozen', False):
        cmd.insert(0, f'"{sys.executable}"')

    logger.info(f'{sys.argv=}')
    logger.info(f'{cmd=}')
    atexit.register(os.execv, sys.executable, cmd)
    if exit_fn:
        exit_fn()
    else:
        sys.exit(0)


def instances_shm():
    return mmap.mmap(-1, 1, 'labbie_instances')


def exit_if_already_running():
    mm = instances_shm()
    if mm[0] > 0:
        mm[0] += 1
        logger.info(f'Labbie is already running, instances={mm[0]}')
        sys.exit()


def make_slotted_dataclass(cls):
    # Need to create a new class, since we can't set __slots__
    #  after a class has been created.

    # Make sure __slots__ isn't already set.
    if '__slots__' in cls.__dict__:
        raise TypeError(f'{cls.__name__} already specifies __slots__')

    # Create a new dict for our new class.
    cls_dict = dict(cls.__dict__)
    field_names = tuple(f.name for f in dataclasses.fields(cls))
    cls_dict['__slots__'] = field_names
    for field_name in field_names:
        # Remove our attributes, if present. They'll still be
        #  available in _MARKER.
        cls_dict.pop(field_name, None)
    # Remove __dict__ itself.
    cls_dict.pop('__dict__', None)
    # And finally create the class.
    qualname = getattr(cls, '__qualname__', None)
    cls = type(cls)(cls.__name__, cls.__bases__, cls_dict)
    if qualname is not None:
        cls.__qualname__ = qualname
    return cls


@contextlib.asynccontextmanager
async def client_session(session: Optional[aiohttp.ClientSession] = None) -> aiohttp.ClientSession:
    if session:
        yield session
    else:
        async with aiohttp.ClientSession() as session:
            yield session


class LogFilter:

    def __init__(self, level):
        self.level = level

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, val):
        self._level = logger.level(val).no

    def __call__(self, record):
        return record['level'].no >= self.level



import datetime
from typing import Any, Dict

import loguru
import injector
import orjson

from labbie import config
from labbie import constants
from labbie import errors
from labbie import ocr
from labbie import result
from labbie import state
from labbie import mods
from labbie.ui import hotkey

logger = loguru.logger
_Config = config.Config
_Constants = constants.Constants
_Result = result.Result
keys = Any  # labbie.ui.keys imports cause circular dependency


@injector.singleton
class AppPresenter:
    @injector.inject
    def __init__(self, constants: _Constants, config: _Config, injector: injector.Injector, app_state: state.AppState, mods: mods.Mods):
        self._constants = constants
        self._config = config
        self._injector = injector
        self._app_state = app_state
        self.mods = mods

        self.presenters = {}

        self._hotkeys: Dict[str, hotkey.Hotkey] = {}
        self._config.ui.hotkeys.attach(self, self._ocr_hotkey_changed, to='ocr')
        self._ocr_hotkey_changed(self._config.ui.hotkeys.ocr)

    def _ocr_hotkey_changed(self, val):
        logger.debug(f'ocr hotkey променен {val=}')
        current_hotkey = self._hotkeys.pop('ocr', None)
        if current_hotkey:
            current_hotkey.stop()

        if val:
            try:
                ocr_hotkey = hotkey.Hotkey(val)
                ocr_hotkey.start(self._ocr_hotkey_pressed)
                self._hotkeys['ocr'] = ocr_hotkey
            except ValueError:
                # May be raised when hotkey can't be bound
                self.show(keys.ErrorWindowKey(Exception(f'Невъзможност за установяване на hotkey: {val}')))

    def _ocr_hotkey_pressed(self):
        self.screen_capture()

    def reset_window_positions(self):
        key = keys.SearchWindowKey()
        if presenter := self.presenters.get(key):
            presenter.reset_position()

    def screen_capture(self):
        if not self._app_state.league_enchants.enabled and not self._app_state.daily_enchants.enabled:
            self.show(keys.ErrorWindowKey('Няма разрешени enchant извличания, моля, редактирайте настройките.'))
            return

        save_path = None
        if self._constants.debug:
            save_path = self._constants.screenshots_dir / datetime.datetime.now().strftime('%Y-%m-%d_%H%M%S')
            save_path.mkdir(exist_ok=True)
        partial_enchant_list = ocr.read_enchants(self._config.ocr.bounds, save_path)
        curr_enchants = self.mods.get_mod_list_from_ocr_results(partial_enchant_list)
        logger.debug(f'{curr_enchants=}')
        if curr_enchants:
            with (self._constants.logs_dir / 'enchants.jsonl').open('ab') as f:
                f.write(orjson.dumps({'timestamp': str(datetime.datetime.utcnow()), 'enchants': curr_enchants})+b'\n')
        results = []
        # TODO: make this work for gloves/boots?
        for index, enchant in enumerate(curr_enchants, start=1):
            logger.debug(f'{index=}: {enchant=}')
            league_matches = None
            if self._app_state.league_enchants.enabled:
                try:
                    league_matches = self._app_state.league_enchants.find_matching_enchants(enchant)
                except errors.EnchantsNotLoaded:
                    self.show(keys.ErrorWindowKey('Enchants не са завършили зареждането.'))
                    return

            daily_matches = None
            if self._app_state.daily_enchants.enabled:
                try:
                    daily_matches = self._app_state.daily_enchants.find_matching_enchants(enchant)
                except errors.EnchantsNotLoaded:
                    self.show(keys.ErrorWindowKey('Enchants не са завършили зареждането.'))
                    return

            if league_matches is not None or daily_matches is not None:
                result = _Result(title=enchant, search=enchant, league_result=league_matches, daily_result=daily_matches, base=False)
                results.append(result)

        key = keys.SearchWindowKey(results, clear=self._config.ocr.clear_previous)
        self.show(key)

    def show(self, key: 'keys._Key'):
        if not isinstance(key, keys._Key):
            raise ValueError(f'{self.__class__.__name__} can only show keys which are instances '
                             f'of WindowKey.')
        if presenter := self.presenters.get(key):
            logger.debug(f'Reusing presenter for {key=}')
            key.show(presenter)
        else:
            logger.debug(f'Getting new presenter for {key=}')
            presenter = key.get_presenter(self._injector)
            self.presenters[key] = presenter
            if key.DELETE_WHEN_CLOSED:
                presenter.add_close_callback(lambda: self.delete(key))
            key.show(presenter)

    def toggle(self, key: 'keys._Key'):
        if not isinstance(key, keys._Key):
            raise ValueError(f'{self.__class__.__name__} can only show keys which are instances '
                             f'of WindowKey.')
        if presenter := self.presenters.get(key):
            logger.debug(f'Reusing presenter for {key=}')
            key.toggle(presenter)
        else:
            self.show(key)

    def delete(self, key):
        logger.debug(f'deleting {key=}')
        self.presenters.pop(key, None)

    def launch(self):
        self.show(keys.SystemTrayIconKey())
        self.show(keys.SearchWindowKey())

    def foreground(self):
        self.show(keys.SearchWindowKey())


from labbie.ui import keys  # noqa: E402



import os

from PyQt5 import QtCore

from labbie.ui import base
from labbie.ui.settings.widget import view as settings_widget


class SettingsWindow(base.BaseWindow):

    def __init__(self, widget: settings_widget.SettingsWidget):
        super().__init__(widget=widget)
        self.set_buttons(minimize=True, close=True)

        flags = self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint
        if os.name == 'posix':
            flags = flags | QtCore.Qt.Tool
        self.setWindowFlags(flags)



import base64
import collections
import dataclasses
import json
from urllib import parse
import webbrowser
from typing import List, Optional

import injector
import loguru

from labbie import bases
from labbie import constants
from labbie import enchants
from labbie import mods
from labbie import result
from labbie.ui.result.widget import view

logger = loguru.logger
_HTML_PAYLOAD_FORMAT = '''<html>
<body>
<h1>Trade Rate Limit Protection</h1>
<h3>Ще бъдете пренасочени към търсенето на {search} след <span id="seconds"></span> секунди.</h3>
С <strong>риск да бъдете временно ограничени</strong>,може да кликнете <a href="{url}">тук</a> to proceed now.

<style>
body {{
    background-color: rgb(53, 53, 53);
    color: rgb(180, 180, 180);
}}
a {{
    color: rgb(56, 252, 196);
}}
</style>
<script>
    var delay = {delay};

    function decrement() {{
        if (delay <= 0) {{
            window.location = "{url}";
            return;
        }}

        document.getElementById('seconds').innerText = Math.floor(delay);
        delay -= .25;
        setTimeout(decrement, 250);
    }}

    decrement();
</script>
</body>
</html>
'''
_MAX_ILVL = 100
_MAX_DISPLAY_ILVL = 86


@dataclasses.dataclass
class ResultData:
    name: str
    base: str
    unique: bool
    ilvl: Optional[int]
    influence: Optional[str]

    def __str__(self):
        ilvl = f'i{self.ilvl}+ ' if self.ilvl else ''
        influence = f'{self.influence} ' if self.influence and self.influence != 'Uninfluenced' else ''
        base = self.base
        return f'{ilvl}{influence}{base}'

    def build_search_url(self, enchant_stat_id, enchant_stat_value):
        stats = [
            {
                'type': 'and',
                'filters': [{'id': enchant_stat_id}]
            }
        ]
        if enchant_stat_value is not None:
            stats[0]['filters'][0].update(
                min=enchant_stat_value,
                max=enchant_stat_value
            )

        query = {
            'status': {
                'option': 'online'
            },
            'stats': stats,
            'type': self.base
        }

        search = {
            'query': query,
            'sort': {
                'price': 'asc'
            }
        }

        query_filters = {}

        if self.unique:
            query['name'] = self.name
        else:
            query_filters['type_filters'] = {
                'filters': {
                    'rarity': {
                        'option': 'nonunique'
                    }
                }
            }

        if self.ilvl:
            query_filters['misc_filters'] = {
                'filters': {
                    'ilvl': {
                        'min': self.ilvl
                    }
                }
            }

        if self.influence and self.influence != 'Uninfluenced':
            influences = [inf.lower() for inf in self.influence.split(', ')]
            stats[0]['filters'].extend({'id': f'pseudo.pseudo_has_{influence}_influence'} for influence in influences)

        if query_filters:
            query['filters'] = query_filters

        query_string = parse.urlencode({'q': json.dumps(search)}, quote_via=parse.quote)
        return f'https://www.pathofexile.com/trade/search/Sentinel?{query_string}'

    def price_check_url(self, enchant_stat_id, enchant_value, delay=0):
        search_url = self.build_search_url(enchant_stat_id, enchant_value)
        logger.debug(f'{search_url=}')
        html = _HTML_PAYLOAD_FORMAT.format(delay=delay, url=search_url, search=self)
        payload = base64.b64encode(html.encode('utf8'))
        return 'https://labbie-redirect.azurewebsites.net/' + payload.decode('utf8')


class ResultWidgetPresenter:

    @injector.inject
    def __init__(self, constants_: constants.Constants, bases_: bases.Bases, mods_: mods.Mods, view: view.ResultWidget):
        self._constants = constants_
        self._bases = bases_
        self._mods = mods_
        self._view = view

        self._mod = None

        # NOTE: connecting the clicked signal to self.on_price_check directly as the slot wasn't working
        # for some reason, no clue.
        self._view.set_price_check_handler(lambda: self.on_price_check())

        self._show_hints = None

    @property
    def widget(self):
        return self._view

    def populate_view(self, result: result.Result):
        self._mod = result.search

        league_results = None
        if result.league_result is not None:
            if result.base:
                league_results = self._build_base_search_display_results(result.league_result)
            else:
                league_results = self._build_enchant_search_display_results(result.league_result)

        daily_results = None
        if result.daily_result is not None:
            if result.base:
                daily_results = self._build_base_search_display_results(result.daily_result)
            else:
                daily_results = self._build_enchant_search_display_results(result.daily_result)

        result_str = result.league_summary(result.base) or ''
        if result_str:
            result_str += '\n\n'
        result_str += result.daily_summary(result.base)

        if result.base:
            self._view.set_price_check_visible(False)
        else:
            self._show_hints = not (self._constants.data_dir / 'result_hints_shown').exists()

        self._view.set_results(
            result.search,
            league_results,
            daily_results,
            results_str=result_str,
            selection_changed_handler=self.on_selection_changed
        )

    def _build_influence_context_menu_items(self, base: str, enchants: List[enchants.Enchant],
                                            ilvl: Optional[int] = None):
        grouped = collections.defaultdict(list)
        for enchant in enchants:
            grouped[', '.join(enchant.influences) or 'Uninfluenced'].append(enchant)

        context_menu_items = []
        sorted_groups = sorted(grouped.items(), key=lambda e: len(e[1]), reverse=True)
        show_all_extras = []
        for influence, influence_enchants in sorted_groups:
            if len(influence_enchants) > 10:
                # Show explicit context menu items above some threshold
                sub_context_menu_items = None
                if not ilvl:
                    sub_context_menu_items = self._build_ilvl_context_menu_items(
                        base, influence_enchants, influence=influence)

                count = len(influence_enchants)
                display_result = view.DisplayResult(
                    count=count,
                    text=influence,
                    data=ResultData(name=base, base=base, unique=False, ilvl=ilvl, influence=influence),
                    indent_level=2 if ilvl else 1,
                    context_menu_items=sub_context_menu_items
                )

                context_menu_items.append(
                    view.ContextMenuItem(
                        section='Influence',
                        text=f'{len(influence_enchants):>5} {influence}',
                        display=display_result
                    )
                )

                if sub_context_menu_items:
                    parent = context_menu_items[-1]
                    for item in sub_context_menu_items:
                        item.set_parent(parent)
            else:
                # Collect the rest into a "Show all" context menu item
                sub_context_menu_items = None
                if not ilvl:
                    sub_context_menu_items = self._build_ilvl_context_menu_items(
                        base, influence_enchants, influence=influence)

                count = len(influence_enchants)
                display_result = view.DisplayResult(
                    count=count,
                    text=influence,
                    data=ResultData(name=base, base=base, unique=False, ilvl=ilvl, influence=influence),
                    indent_level=2 if ilvl else 1,
                    context_menu_items=sub_context_menu_items
                )
                show_all_extras.append(display_result)

        if show_all_extras:
            context_menu_items.append(
                view.ContextMenuItem(
                    section='Influence',
                    text='Show All',
                    display=context_menu_items + show_all_extras
                )
            )

            for display_result in show_all_extras:
                if display_result.context_menu_items:
                    parent = context_menu_items[-1]
                    for item in display_result.context_menu_items:
                        item.set_parent(parent)

        for index, item in enumerate(context_menu_items):
            item.index = index

        return context_menu_items

    def _build_ilvl_context_menu_items(self, base: str, enchants: List[enchants.Enchant],
                                       influence: Optional[str] = None):
        total = len(enchants)
        grouped = collections.defaultdict(list)
        for enchant in enchants:
            grouped[enchant.ilvl].append(enchant)

        ilvl_count = len(grouped)
        context_menu_items = []

        covered_enchants = []
        for ilvl in range(_MAX_ILVL, _MAX_DISPLAY_ILVL - 1, -1):
            if ilvl_enchants := grouped.get(ilvl):
                covered_enchants.extend(ilvl_enchants)
                ilvl_count -= 1  # combining ilvls about max into max reduces count

        if covered_enchants:
            sub_context_menu_items = None
            if not influence:
                sub_context_menu_items = self._build_influence_context_menu_items(
                    base, covered_enchants, ilvl=ilvl)
            count = len(covered_enchants)
            context_menu_items.append(
                view.ContextMenuItem(
                    section='Item Level',
                    text=f'{count / total * 100:>3.0f}% i{_MAX_DISPLAY_ILVL}+',
                    display=view.DisplayResult(
                        count=count,
                        text=f'i{_MAX_DISPLAY_ILVL}+',
                        data=ResultData(name=base, base=base, unique=False, ilvl=_MAX_DISPLAY_ILVL, influence=influence),
                        indent_level=2 if influence else 1,
                        context_menu_items=sub_context_menu_items
                    )
                )
            )

            if sub_context_menu_items:
                parent = context_menu_items[-1]
                for item in sub_context_menu_items:
                    item.set_parent(parent)

        cumulative_total = len(covered_enchants)
        for ilvl in range(_MAX_DISPLAY_ILVL - 1, 0, -1):
            if (ilvl_enchants := grouped.get(ilvl)) is None:
                continue

            if len(context_menu_items) == 5:
                break
            elif len(context_menu_items) == ilvl_count:
                break

            cumulative_total += len(ilvl_enchants)
            covered_enchants.extend(ilvl_enchants)

            sub_context_menu_items = None
            if not influence:
                sub_context_menu_items = self._build_influence_context_menu_items(
                    base, covered_enchants, ilvl=ilvl)

            context_menu_items.append(
                view.ContextMenuItem(
                    section='Item Level',
                    text=f'{cumulative_total / total * 100:>3.0f}% i{ilvl}+',
                    display=view.DisplayResult(
                        count=cumulative_total,
                        text=f'i{ilvl}+',
                        data=ResultData(name=base, base=base, unique=False, ilvl=ilvl, influence=influence),
                        indent_level=2 if influence else 1,
                        context_menu_items=sub_context_menu_items
                    )
                )
            )

            if sub_context_menu_items:
                parent = context_menu_items[-1]
                for item in sub_context_menu_items:
                    item.set_parent(parent)

        if len(context_menu_items) < ilvl_count:
            show_all_extras = []
            for ilvl in range(ilvl, 0, -1):
                if (ilvl_enchants := grouped.get(ilvl)) is None:
                    continue

                cumulative_total += len(ilvl_enchants)
                covered_enchants.extend(ilvl_enchants)

                sub_context_menu_items = None
                if not influence:
                    sub_context_menu_items = self._build_influence_context_menu_items(
                        base, covered_enchants, ilvl=ilvl)

                show_all_extras.append(
                    view.DisplayResult(
                        count=cumulative_total,
                        text=f'i{ilvl}+',
                        data=ResultData(name=base, base=base, unique=False, ilvl=ilvl, influence=influence),
                        indent_level=2 if influence else 1,
                        context_menu_items=sub_context_menu_items
                    )
                )

            context_menu_items.append(
                view.ContextMenuItem(
                    section='Item Level',
                    text='Show All',
                    display=context_menu_items + show_all_extras
                )
            )

            for display_result in show_all_extras:
                if display_result.context_menu_items:
                    parent = context_menu_items[-1]
                    for item in display_result.context_menu_items:
                        item.set_parent(parent)

        for index, item in enumerate(context_menu_items):
            item.index = index

        return context_menu_items

    def _build_context_menu_items(self, base: str, enchants: List[enchants.Enchant]):
        context_menu_items = []
        context_menu_items.extend(self._build_influence_context_menu_items(base, enchants))
        context_menu_items.extend(self._build_ilvl_context_menu_items(base, enchants))

        for index, item in enumerate(context_menu_items):
            item.index = index

        return context_menu_items

    def _build_base_search_display_results(self, results: List[enchants.Enchant]):
        enchants = collections.Counter()

        for enchant in results:
            for mod in enchant.mods:
                if mod not in self._mods.helm_mods:
                    continue
                enchants[mod] += 1

        display_results = []
        for index, (enchant, count) in enumerate(enchants.most_common()):
            display_result = view.DisplayResult(
                count=count,
                text=enchant,
                data=None
            )
            display_result.index = index  # this is not a constructor arg, needs to be set here
            display_results.append(display_result)

        return (len(results), display_results)

    def on_price_check(self):
        mod_info = self._mods.helm_mod_info.get(self._mod)
        if mod_info is None:
            logger.error(f'Невъзможно е да се намери информация за "{self._mod}"')
            return

        if mod_info.trade_stat_id is None:
            logger.error(f'Няма trade идентификатор за "{self._mod}" {mod_info=}')
            return

        selected: List[view.DisplayResult] = self._view.get_selected_data()
        delay = 0
        for index, result in enumerate(selected):
            delay = index // 3 * 5.25
            webbrowser.open_new_tab(result.data.price_check_url(mod_info.trade_stat_id, mod_info.trade_stat_value, delay))

    def _build_enchant_search_display_results(self, results: List[enchants.Enchant]):
        all_bases = collections.Counter()
        rare_bases = collections.defaultdict(list)

        for enchant in results:
            all_bases[enchant.display_name] += 1
            if not enchant.unique:
                rare_bases[enchant.item_base].append(enchant)

        display_results = []
        for index, (base, count) in enumerate(all_bases.most_common()):
            context_menu_items = None
            if enchants := rare_bases.get(base):
                context_menu_items = self._build_context_menu_items(base, enchants)

            unique = base not in rare_bases
            # krangle the base so that uniques get the actual base type here
            krangled_base = self._bases.helms[base].base if unique else base
            display_result = view.DisplayResult(
                count=count,
                text=base,
                data=ResultData(name=base, base=krangled_base, unique=unique, ilvl=None, influence=None),
                context_menu_items=context_menu_items
            )
            display_result.index = index  # this is not a constructor arg, needs to be set here
            display_results.append(display_result)

        return (len(results), display_results)

    def on_price_check(self):
        mod_info = self._mods.helm_mod_info.get(self._mod)
        if mod_info is None:
            logger.error(f'Невъзможно е да се намери информация за "{self._mod}"')
            return

        if mod_info.trade_stat_id is None:
            logger.error(f'Няма trade идентификатор за "{self._mod}" {mod_info=}')
            return

        selected: List[view.DisplayResult] = self._view.get_selected_data()
        delay = 0
        for index, result in enumerate(selected):
            delay = index // 3 * 5.25
            webbrowser.open_new_tab(result.data.price_check_url(mod_info.trade_stat_id, mod_info.trade_stat_value, delay))

    def on_selection_changed(self):
        if self._show_hints:
            self._view.show_right_click_hint()
            (self._constants.data_dir / 'result_hints_shown').touch()
            self._show_hints = False
        # selected = self._view.get_selected_data()

        # if not selected:
        #     self._view.set_selected_stats_text('')
        #     return

        # selected_results = set()
        # max_selected_subresult_index = collections.defaultdict(int)

        # for item in selected:
        #     data = item.data(Qt.UserRole)
        #     if data.parent_index is None:
        #         selected_results.add(data.index)
        #     else:
        #         cur_max = max_selected_subresult_index[data.parent_index]
        #         max_selected_subresult_index[data.parent_index] = max(cur_max, data.index)

        # selected_count = 0
        # for index in selected_results:
        #     max_selected_subresult_index.pop(index, None)
        #     selected_count += self._current_results[index].count

        # partial_selections = []
        # for parent_index, subresult_index in max_selected_subresult_index.items():
        #     parent = self._current_results[parent_index]
        #     count = parent.subresult(subresult_index).count
        #     selected_count += count
        #     partial_selections.append((parent.text, count, parent.count))

        # if self._current_results is self._league_results:
        #     total = self._total_league_results
        # else:
        #     total = self._total_daily_results

        # text = [f'<strong>Total:</strong> {selected_count / total * 100:0.2f}% ({selected_count}/{total})']
        # for partial_text, partial_count, partial_total in partial_selections:
        #     text.append(f'<strong>{partial_text}:</strong> {partial_count / partial_total * 100:0.2f}% '
        #                 f'({partial_count}/{partial_total})')
        # text = '<br />'.join(text)
        # self.lbl_selected_stats.setText(text)

    def show(self):
        self._view.show()



from PyQt5 import QtCore
from PyQt5 import QtWidgets

from labbie.ui import base

Qt = QtCore.Qt


class ErrorWidget(base.BaseWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.lbl_error = QtWidgets.QLabel(self)
        self.lbl_error.setStyleSheet("QLabel{font-size: 12pt;}")
        self.lbl_error.setSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Minimum)
        self.btn_ok = QtWidgets.QPushButton('Добре', self)

        layout = QtWidgets.QVBoxLayout()
        layout.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinimumSize)
        layout.addSpacing(10)
        layout.addWidget(self.lbl_error)
        layout.addSpacing(10)
        layout.addWidget(self.btn_ok, 0, Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight)
        self.setLayout(layout)

        self.btn_ok.clicked.connect(self.close)

        self.setWindowTitle('Грешка')
        self.center_on_screen()

    def set_error(self, error: str):
        self.lbl_error.setText(error)



import asyncio

import injector
import loguru

from labbie import config
from labbie import constants
from labbie import state
from labbie.ui import keys
from labbie.ui.app import presenter as app
from labbie.ui.settings.widget import view
from labbie.ui.screen_selection.widget import view as screen_selection

logger = loguru.logger
_Config = config.Config
_Constants = constants.Constants


class SettingsPresenter:

    @injector.inject
    def __init__(
        self,
        constants: _Constants,
        app_state: state.AppState,
        app_presenter: app.AppPresenter,
        config: _Config,
        view: view.SettingsWidget,
        screen_selection_view_builder: injector.AssistedBuilder[screen_selection.ScreenSelectionWidget]
    ):
        self._constants = constants
        self._app_state = app_state
        self._app_presenter = app_presenter
        self._config = config
        self._view = view
        self._screen_selection_view_builder = screen_selection_view_builder
        self._screen_selection_view = None

        self._view.set_select_bounds_handler(self.on_select_bounds)
        self._view.set_reset_window_positions_handler(self.on_reset_window_positions)
        self._view.set_save_handler(self.on_save)

        self._view.league = self._config.league
        self._view.daily = self._config.daily
        self._view.clear_previous = self._config.ocr.clear_previous
        self._view.hotkey = self._config.ui.hotkeys.ocr
        self._view.left = str(self._config.ocr.bounds.left)
        self._view.top = str(self._config.ocr.bounds.top)
        self._view.right = str(self._config.ocr.bounds.right)
        self._view.bottom = str(self._config.ocr.bounds.bottom)

    @property
    def widget(self):
        return self._view

    def on_reset_window_positions(self):
        self._app_presenter.reset_window_positions()

    def on_select_bounds(self, checked):
        if self._screen_selection_view is None:
            self._screen_selection_view = self._screen_selection_view_builder.build(
                left=int(self._view.left),
                top=int(self._view.top),
                right=int(self._view.right),
                bottom=int(self._view.bottom)
            )
            self._screen_selection_view.set_done_handler(self.on_screen_selection_done)
            self._screen_selection_view.show()
            self._view.set_select_button_text('Задаване')
        else:
            self.on_screen_selection_done()

    def on_screen_selection_done(self):
        left, top, right, bottom = self._screen_selection_view.get_position()
        self._screen_selection_view.hide()
        self._screen_selection_view = None
        self._view.set_select_button_text('Изберете')
        self._view.left = str(left)
        self._view.top = str(top)
        # TODO(bnorick): identify and fix the underlying bug instead of + 1 as a bandaid fix
        self._view.right = str(right + 1)
        self._view.bottom = str(bottom + 1)

    async def on_save(self, checked):
        try:
            left = int(self._view.left)
            top = int(self._view.top)
            right = int(self._view.right)
            bottom = int(self._view.bottom)

            if right <= left:
                raise ValueError('Лявото трябва да е по-малко от дясното')

            if bottom <= top:
                raise ValueError('Горната част трябва да е по-малка от долната')
        except ValueError as e:
            err = str(e)
            if err.startswith('невалидна стойност'):
                e = ValueError('Стойностите на границите (ляво, горе, дясно, долу) трябва да са цели числа')

            self._app_presenter.show(keys.ErrorWindowKey(e))
            return

        if self._view.league != self._config.league:
            self._config.league = self._view.league
            if self._config.league:
                asyncio.create_task(self._app_state.league_enchants.download_or_load(self._constants))
            else:
                self._app_state.league_enchants.set_enchants(None, None)

        if self._view.daily != self._config.daily:
            self._config.daily = self._view.daily
            if self._config.daily:
                asyncio.create_task(self._app_state.daily_enchants.download_or_load(self._constants))
            else:
                self._app_state.daily_enchants.set_enchants(None, None)

        if self._view.hotkey != self._config.ui.hotkeys.ocr:
            self._config.ui.hotkeys.ocr = self._view.hotkey
            self._config.ui.hotkeys.notify(ocr=self._view.hotkey)
        self._config.ocr.clear_previous = self._view.clear_previous
        self._config.ocr.bounds.left = int(self._view.left)
        self._config.ocr.bounds.top = int(self._view.top)
        self._config.ocr.bounds.right = int(self._view.right)
        self._config.ocr.bounds.bottom = int(self._view.bottom)
        self._config.save()
        self._view.close()

    def cleanup(self):
        if self._screen_selection_view:
            self._screen_selection_view.close()

    def show(self):
        self._view.show()



import os
import pathlib
from typing import List, Optional

import cv2 as cv
import loguru
import numpy as np
from PIL import ImageGrab, Image
import pytesseract

from labbie import bounds
from labbie import utils

logger = loguru.logger
_KRANGLES = [
    ('Sammon', 'Summon'),
]

if os.name == 'nt':
    pytesseract.pytesseract.tesseract_cmd = str(utils.bin_dir() / 'tesseract' / 'tesseract.exe')
else:
    pytesseract.pytesseract.tesseract_cmd = 'tesseract'


def read_enchants(bounds_: bounds.Bounds, save_path: Optional[pathlib.Path], dilate: Optional[bool] = False):
    if save_path and not save_path.exists():
        save_path.mkdir(exist_ok=True, parents=True)
    image = ImageGrab.grab(bounds_.as_tuple(), all_screens=True)
    if save_path:
        image.save(save_path / 'full.png')
    enchants = parse_image(image, save_path, dilate)
    return enchants


def parse_image(image, save_path, dilate):
    image = np.array(image)
    grayscale = cv.cvtColor(image, cv.COLOR_RGB2GRAY)

    sobel = cv.Sobel(grayscale, -1, 1, 0)
    mask = np.any(sobel > 245, axis=1)

    # Image.fromarray(test).show()
    # image = cv.medianBlur(image, 5)
    # im_bw = cv.adaptiveThreshold(image, 255, cv.ADAPTIVE_THRESH_MEAN_C, cv.THRESH_BINARY_INV, 25, 3)

    # image = cv.GaussianBlur(image, (5,5), 0)
    # (thresh, im_bw) = cv.threshold(image, 0, 255, cv.THRESH_BINARY + cv.THRESH_OTSU)

    r, g, b = cv.split(image)
    # h = r.shape[0]
    (thresh, im_bw) = cv.threshold(r, 90, 255, cv.THRESH_BINARY_INV)

    # test = np.copy(im_bw)
    # test[~mask, :] = 255
    # Image.fromarray(test).show()

    im_bw[~mask, :] = 255

    # mask = im_bw.sum(axis=1) < .05 * h
    # bad_rows = np.nonzero(mask)[0]
    # mask[bad_rows + 1] = True
    # mask[bad_rows - 1] = True
    # im_bw[mask,:] = 255

    # grayscale = cv.cvtColor(np.array(image), cv.COLOR_RGB2GRAY)
    # (thresh, im_bw) = cv.threshold(grayscale, 0, 255, cv.THRESH_BINARY_INV + cv.THRESH_OTSU)

    if dilate:
        kernel = np.ones((2,2),np.uint8)
        # errosion is equivalent to dilation for white back ground black text.
        im_bw = cv.erode(im_bw, kernel, iterations=1)
    if save_path:
        Image.fromarray(im_bw).save(save_path / 'full_processed.png')
    enchants = pytesseract.image_to_string(im_bw, config='--psm 12').replace('\x0c', '').replace('’', "'")
    enchants = [e.strip().rstrip('.') for e in enchants.split('\n') if e]
    return _fix_krangled_ocr(enchants)


def _fix_krangled_ocr(enchants: List[str]):
    unkrangled_enchants = []
    for enchant in enchants:
        for krangle in _KRANGLES:
            enchant = enchant.replace(*krangle)
        unkrangled_enchants.append(enchant)
    return unkrangled_enchants



import dataclasses
from typing import List, Optional

from labbie import enchants


@dataclasses.dataclass
class Result:
    title: str
    search: str
    base: bool  # is this a result from a base search
    league_result: Optional[List[enchants.Enchant]]
    daily_result: Optional[List[enchants.Enchant]]

    def _summary(self, type_: str, base: bool):
        results = getattr(self, f'{type_}_result')
        if results is None:
            return None
        if base:
            return f'{type_.upper()} ({len(results)})\n{enchants.base_summary(results)}'
        return f'{type_.upper()} ({len(results)})\n{enchants.enchant_summary(results)}'

    def league_summary(self, base):
        return self._summary('league', base)

    def daily_summary(self, base):
        return self._summary('daily', base)

    def __repr__(self):
        if self.league_result:
            league_result_repr = [self.league_result[0]]
            if len(self.league_result) > 1:
                league_result_repr.append(f'(... {len(self.league_result) - 1} other results ...)')
        else:
            league_result_repr = self.league_result

        if self.daily_result:
            daily_result_repr = [self.daily_result[0]]
            if len(self.daily_result) > 1:
                daily_result_repr.append(f'(... {len(self.daily_result) - 1} other results ...)')
        else:
            daily_result_repr = self.daily_result

        return f'{type(self).__name__}(title={self.title!r}, search={self.search!r}, league_result={daily_result_repr!r}, daily_result={daily_result_repr!r})'


from PyQt5 import QtCore
from PyQt5 import QtWidgets

Qt = QtCore.Qt

# TODO(bnorick): make matching work similar to ~ on official trade site
# TODO(bnorick): make mouseover highlight after matching

# From https://stackoverflow.com/a/4829759
# Licensed under CC-BY-SA 3.0 (https://creativecommons.org/licenses/by-sa/3.0/)

class FuzzyComboBox(QtWidgets.QComboBox):

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setEditable(True)

        # add a filter model to filter matching items
        self.pFilterModel = QtCore.QSortFilterProxyModel(self)
        self.pFilterModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.pFilterModel.setSourceModel(self.model())

        # add a completer, which uses the filter model
        self.completer = QtWidgets.QCompleter(self.pFilterModel, self)
        # always show all (filtered) completions
        self.completer.setCompletionMode(QtWidgets.QCompleter.UnfilteredPopupCompletion)
        self.setCompleter(self.completer)

        # connect signals
        self.lineEdit().textEdited.connect(self.pFilterModel.setFilterFixedString)
        self.completer.activated.connect(self.on_completer_activated)

    # on selection of an item from the completer, select the corresponding item from combobox
    def on_completer_activated(self, text):
        if text:
            index = self.findText(text)
            self.setCurrentIndex(index)
            self.activated[str].emit(self.itemText(index))

    # on model change, update the models of the filter and completer as well
    def setModel(self, model):
        super().setModel(model)
        self.pFilterModel.setSourceModel(model)
        self.completer.setModel(self.pFilterModel)

    # on model column change, update the model column of the filter and completer as well
    def setModelColumn(self, column):
        self.completer.setCompletionColumn(column)
        self.pFilterModel.setFilterKeyColumn(column)
        super().setModelColumn(column)



from PyQt5.QtGui import QPalette, QColor
from ._utils import QT_VERSION, resource_path

_STYLESHEET = resource_path('resources/style.qss')
""" str: Main stylesheet. """


def _apply_base_theme(app):
    """ Apply base theme to the application.

        Args:
            app (QApplication): QApplication instance.
    """

    if QT_VERSION < (5,):
        app.setStyle('plastique')
    else:
        app.setStyle('Fusion')

    with open(_STYLESHEET) as stylesheet:
        app.setStyleSheet(stylesheet.read())


def dark(app):
    """ Apply Dark Theme to the Qt application instance.

        Args:
            app (QApplication): QApplication instance.
    """

    darkPalette = QPalette()

    # base
    darkPalette.setColor(QPalette.WindowText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Button, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.Light, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Midlight, QColor(90, 90, 90))
    darkPalette.setColor(QPalette.Dark, QColor(35, 35, 35))
    darkPalette.setColor(QPalette.Text, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.BrightText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.ButtonText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Base, QColor(42, 42, 42))
    darkPalette.setColor(QPalette.Window, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    darkPalette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    darkPalette.setColor(QPalette.HighlightedText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Link, QColor(56, 252, 196))
    darkPalette.setColor(QPalette.AlternateBase, QColor(66, 66, 66))
    darkPalette.setColor(QPalette.ToolTipBase, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.ToolTipText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.LinkVisited, QColor(80, 80, 80))

    # disabled
    darkPalette.setColor(QPalette.Disabled, QPalette.WindowText,
                         QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.Text,
                         QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.ButtonText,
                         QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.Highlight,
                         QColor(80, 80, 80))
    darkPalette.setColor(QPalette.Disabled, QPalette.HighlightedText,
                         QColor(127, 127, 127))

    app.setPalette(darkPalette)
    
    _apply_base_theme(app)


def light(app):
    """ Apply Light Theme to the Qt application instance.

        Args:
            app (QApplication): QApplication instance.
    """

    lightPalette = QPalette()

    # base
    lightPalette.setColor(QPalette.WindowText, QColor(0, 0, 0))
    lightPalette.setColor(QPalette.Button, QColor(240, 240, 240))
    lightPalette.setColor(QPalette.Light, QColor(180, 180, 180))
    lightPalette.setColor(QPalette.Midlight, QColor(200, 200, 200))
    lightPalette.setColor(QPalette.Dark, QColor(225, 225, 225))
    lightPalette.setColor(QPalette.Text, QColor(0, 0, 0))
    lightPalette.setColor(QPalette.BrightText, QColor(0, 0, 0))
    lightPalette.setColor(QPalette.ButtonText, QColor(0, 0, 0))
    lightPalette.setColor(QPalette.Base, QColor(237, 237, 237))
    lightPalette.setColor(QPalette.Window, QColor(240, 240, 240))
    lightPalette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    lightPalette.setColor(QPalette.Highlight, QColor(76, 163, 224))
    lightPalette.setColor(QPalette.HighlightedText, QColor(0, 0, 0))
    lightPalette.setColor(QPalette.Link, QColor(0, 162, 232))
    lightPalette.setColor(QPalette.AlternateBase, QColor(225, 225, 225))
    lightPalette.setColor(QPalette.ToolTipBase, QColor(240, 240, 240))
    lightPalette.setColor(QPalette.ToolTipText, QColor(0, 0, 0))
    lightPalette.setColor(QPalette.LinkVisited, QColor(222, 222, 222))

    # disabled
    lightPalette.setColor(QPalette.Disabled, QPalette.WindowText,
                         QColor(115, 115, 115))
    lightPalette.setColor(QPalette.Disabled, QPalette.Text,
                         QColor(115, 115, 115))
    lightPalette.setColor(QPalette.Disabled, QPalette.ButtonText,
                         QColor(115, 115, 115))
    lightPalette.setColor(QPalette.Disabled, QPalette.Highlight,
                         QColor(190, 190, 190))
    lightPalette.setColor(QPalette.Disabled, QPalette.HighlightedText,
                         QColor(115, 115, 115))

    app.setPalette(lightPalette)

    _apply_base_theme(app)
    



import abc
import asyncio
import dataclasses
import inspect
from typing import ClassVar, List, Union

import injector

from labbie import result
from labbie.ui.about.window import presenter as about
from labbie.ui.system_tray import presenter as system_tray
from labbie.ui.search.window import presenter as search
from labbie.ui.settings.window import presenter as settings
from labbie.ui.error.window import presenter as error


@dataclasses.dataclass(frozen=True)
class _Key(abc.ABC):
    DELETE_WHEN_CLOSED: ClassVar[bool] = False

    @abc.abstractmethod
    def get_presenter(self, injector: injector.Injector):
        raise NotImplementedError

    def show(self, presenter):
        presenter.show()

    def toggle(self, presenter):
        presenter.toggle()


@dataclasses.dataclass(frozen=True)
class _PopulatableKey(_Key):
    @abc.abstractmethod
    def _populate_presenter(self, presenter):
        raise NotImplementedError

    def show(self, presenter):
        populate_result = self._populate_presenter(presenter)
        if inspect.iscoroutine(populate_result):
            asyncio.create_task(self._show_after_populate(presenter, populate_result))
        else:
            presenter.show()

    async def _show_after_populate(self, presenter, populate_task):
        await populate_task
        presenter.show()


@dataclasses.dataclass(frozen=True)
class SystemTrayIconKey(_Key):
    def get_presenter(self, injector: injector.Injector):
        return injector.get(system_tray.SystemTrayIconPresenter)


@dataclasses.dataclass(frozen=True)
class SearchWindowKey(_PopulatableKey):
    DELETE_WHEN_CLOSED: ClassVar[bool] = False
    results: Union[None, result.Result, List[result.Result]] = dataclasses.field(default=None, compare=False)
    clear: bool = dataclasses.field(default=False, compare=False)

    def get_presenter(self, injector: injector.Injector):
        return injector.get(search.SearchWindowPresenter)

    def _populate_presenter(self, presenter: search.SearchWindowPresenter):
        presenter.populate_view(self.results, clear=self.clear)


@dataclasses.dataclass(frozen=True)
class ErrorWindowKey(_PopulatableKey):
    DELETE_WHEN_CLOSED: ClassVar[bool] = True
    exception: Exception = dataclasses.field(compare=False)

    def get_presenter(self, injector: injector.Injector):
        return injector.get(error.ErrorWindowPresenter)

    def _populate_presenter(self, presenter: error.ErrorWindowPresenter):
        presenter.populate_view(self.exception)


@dataclasses.dataclass(frozen=True)
class SettingsWindowKey(_Key):
    DELETE_WHEN_CLOSED: ClassVar[bool] = True

    def get_presenter(self, injector: injector.Injector):
        return injector.get(settings.SettingsWindowPresenter)


@dataclasses.dataclass(frozen=True)
class AboutWindowKey(_Key):
    DELETE_WHEN_CLOSED: ClassVar[bool] = True

    def get_presenter(self, injector: injector.Injector):
        return injector.get(about.AboutWindowPresenter)



import collections
import dataclasses
import functools
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import injector
import pyperclip
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from labbie.ui import base
from labbie.ui import clickable_label
from labbie.ui import switch
from labbie.ui import transparent_window
from labbie.ui import utils

Qt = QtCore.Qt
_LEAGUE_FORMAT = 'Лига ({})'
_DAILY_FORMAT = 'Дневно ({})'
_INDENT = 25


@dataclasses.dataclass
class DisplayResult:
    index: Optional[int] = dataclasses.field(init=False, default=None)
    count: int
    text: str
    data: Any
    indent_level: int = 0
    context_menu_items: Optional[List['ContextMenuItem']] = None

    _displayed_context_menu_indices: Set[int] = dataclasses.field(init=False, default_factory=set)

    def __hash__(self):
        return id(self)

    def context_menu_results_displayed_before(self, index: int):
        return sum(self.context_menu_items[displayed_index]._displayed_result_count
                   for displayed_index in self._displayed_context_menu_indices
                   if displayed_index < index)

    def get_displayable_context_menu_items(self) -> Dict[str, List[Tuple[int, 'ContextMenuItem']]]:
        if not self.context_menu_items:
            return {}

        sections = collections.defaultdict(list)
        for item in self.context_menu_items:
            if item.index in self._displayed_context_menu_indices:
                continue

            sections[item.section].append(item)

        return sections


@dataclasses.dataclass
class ContextMenuItem:
    index: int = dataclasses.field(init=False)
    _displayed_result_count: int = dataclasses.field(init=False, default=0)
    _parent: Optional['ContextMenuItem'] = dataclasses.field(init=False, default=None)
    section: str
    text: str
    display: Union[DisplayResult, List[Union[DisplayResult, 'ContextMenuItem']]]

    def increment_displayed_results(self):
        self._displayed_result_count += 1
        parent = self._parent
        while parent:
            parent._displayed_result_count += 1
            parent = parent._parent

    def set_parent(self, parent: 'ContextMenuItem'):
        self._parent = parent


class ResultWidget(base.BaseWidget):

    @injector.inject
    def __init__(self, parent=None):
        super().__init__(parent)

        self._league_results = None
        self._daily_results = None
        self._current_results = None
        self._results_str = ''

        self.widget_hint = None

        self.widget_type = QtWidgets.QWidget(self)
        self.toggle_type = switch.Toggle(self.widget_type, thumb_radius=8, track_radius=5)
        self.lbl_league = clickable_label.ClickableLabel(self.widget_type)
        self.lbl_daily = clickable_label.ClickableLabel(self.widget_type)
        self.lbl_title = QtWidgets.QLabel(self)

        self.btn_price_check = QtWidgets.QPushButton('Проверка на цената', self)
        self.btn_copy = QtWidgets.QPushButton('Копиране на резултатите', self)
        self.lbl_selected_stats = QtWidgets.QLabel(self)
        self.btn_close = QtWidgets.QPushButton('Затвори', self)

        self.stack_results = QtWidgets.QStackedWidget(self)
        self._widget_league_results = None
        self._widget_daily_results = None

        self.lbl_league.setStyleSheet('QLabel { font-weight: bold; }')
        self.lbl_daily.setStyleSheet('QLabel { font-weight: bold; }')
        self.lbl_league.hide()
        self.toggle_type.hide()
        self.lbl_daily.hide()
        self.btn_price_check.setEnabled(False)  # enabled upon selection

        layout_type = QtWidgets.QHBoxLayout()
        layout_type.addWidget(self.lbl_league)
        layout_type.addWidget(self.toggle_type)
        layout_type.addWidget(self.lbl_daily)
        layout_type.addSpacing(15)
        layout_type.setContentsMargins(0, 0, 0, 0)

        self.widget_type.setLayout(layout_type)

        layout_top = QtWidgets.QHBoxLayout()
        layout_top.addWidget(self.widget_type)
        layout_top.addWidget(self.lbl_title, 1)

        layout_right = QtWidgets.QVBoxLayout()
        layout_right.addWidget(self.btn_price_check)
        layout_right.addWidget(self.btn_copy)
        layout_right.addWidget(self.lbl_selected_stats, 1)
        layout_right.addWidget(self.btn_close)

        layout = QtWidgets.QGridLayout()
        layout.addLayout(layout_top, 0, 0, 1, 2)
        layout.addWidget(self.stack_results, 1, 0)
        layout.addLayout(layout_right, 1, 1)
        self.setLayout(layout)

        self.toggle_type.toggled.connect(self._on_type_toggled)
        self.lbl_league.clicked.connect(lambda: self.toggle_type.setChecked(False))
        self.lbl_daily.clicked.connect(lambda: self.toggle_type.setChecked(True))
        self.btn_copy.clicked.connect(lambda _: pyperclip.copy(self._results_str))

        def close_self():
            tab_widget = self.parentWidget().parentWidget()
            index = tab_widget.indexOf(self)
            tab_widget.removeTab(index)

        self.btn_close.clicked.connect(close_self)

    def _on_results_selection_changed(self):
        selected = self.stack_results.currentWidget().selectedItems()
        items_selected = bool(selected)
        self.btn_price_check.setEnabled(items_selected)
        # self._update_selected_stats(selected)

    def _on_type_toggled(self, is_daily):
        self._set_active_type(is_daily)
        self._current_results = self._daily_results if is_daily else self._league_results
        widget = self._widget_daily_results if is_daily else self._widget_league_results
        self.stack_results.setCurrentWidget(widget)
        selected = widget.selectedItems()
        items_selected = bool(selected)
        self.btn_price_check.setEnabled(items_selected)
        # self._update_selected_stats(selected)

    def _set_active_type(self, is_daily):
        palette = self.palette()
        inactive = palette.color(palette.ColorGroup.Disabled, palette.ColorRole.Text).name()
        active = palette.text().color().name()
        if is_daily:
            self.lbl_league.setStyleSheet(f'QLabel {{ color: {inactive}; font-weight:bold; }}')
            self.lbl_daily.setStyleSheet(f'QLabel {{ color: {active}; font-weight:bold; }}')
        else:
            self.lbl_league.setStyleSheet(f'QLabel {{ color: {active}; font-weight:bold; }}')
            self.lbl_daily.setStyleSheet(f'QLabel {{ color: {inactive}; font-weight:bold; }}')

    def _show_context_menu(self, point: QtCore.QPoint):
        list_results = self.stack_results.currentWidget()
        item = list_results.itemAt(point)
        row = list_results.row(item)
        result: DisplayResult = item.data(Qt.UserRole)
        menu_items = result.get_displayable_context_menu_items()

        if not menu_items:
            return

        self.hide_hints()

        menu = QtWidgets.QMenu(self)
        for section, items in menu_items.items():
            menu.addSection(section)
            for menu_item in items:
                action = menu.addAction(menu_item.text)
                action.triggered.connect(
                    functools.partial(
                        self._add_context_menu_result,
                        source=result,
                        source_row=row,
                        context_menu_index=menu_item.index
                    )
                )
        menu.exec(list_results.mapToGlobal(point))

    def _add_context_menu_result(self, source: DisplayResult, source_row: int, context_menu_index: int):
        context_menu_item = source.context_menu_items[context_menu_index]
        displayed_before = source.context_menu_results_displayed_before(context_menu_index)
        row = source_row + 1 + displayed_before

        if source.indent_level == 0 and not source._displayed_context_menu_indices:
            self._add_space_to_list(index=row)

        if isinstance(context_menu_item.display, DisplayResult):
            self._add_result_to_list(context_menu_item.display, index=row)
            context_menu_item.increment_displayed_results()
        else:
            for result in reversed(context_menu_item.display):
                if isinstance(result, DisplayResult):
                    self._add_result_to_list(result, index=row)
                    context_menu_item.increment_displayed_results()
                else:
                    if result.index not in source._displayed_context_menu_indices:
                        displayed_before = source.context_menu_results_displayed_before(result.index)
                        row = source_row + 1 + displayed_before
                        self._add_result_to_list(result.display, index=row)
                        context_menu_item.increment_displayed_results()
                        source._displayed_context_menu_indices.add(result.index)

        source._displayed_context_menu_indices.add(context_menu_index)

    def show_right_click_hint(self):
        if self.widget_hint:
            self.widget_hint.show()
        else:
            widget = transparent_window.TransparentWindow(parent=self, close_on_click=True)

            pixmap = utils.recolored_icon('info.svg', 83).pixmap(20, 20)
            icon = QtWidgets.QLabel(widget)
            icon.setPixmap(pixmap)

            label = QtWidgets.QLabel('Десен клик върху rare bases\nпоказва допълнителни филтри', widget)
            label.setStyleSheet('QLabel { color: rgb(37, 37, 37); font-size: 10pt; }')

            layout = QtWidgets.QHBoxLayout()
            layout.addWidget(icon, 0)
            layout.addWidget(label)
            widget.setLayout(layout)

            widget.setMinimumWidth(130)
            widget.adjustSize()

            results_geom = self.stack_results.geometry()
            delta = results_geom.topRight() - widget.geometry().topRight()
            widget.move(delta - QtCore.QPoint(50, -12))
            widget.show()

            self.widget_hint = widget

    def hide_hints(self):
        if self.widget_hint:
            self.widget_hint.close()
            self.widget_hint = None


    # def _update_selected_stats(self, selected: List[QtWidgets.QListWidgetItem] = None):
    #     if selected is None:
    #         selected = self.stack_results.currentWidget().selectedItems()

    #     if not selected:
    #         self.lbl_selected_stats.setText('')
    #         return

    #     selected_results = set()
    #     max_selected_subresult_index = collections.defaultdict(int)

    #     for item in selected:
    #         data = item.data(Qt.UserRole)
    #         if data.parent_index is None:
    #             selected_results.add(data.index)
    #         else:
    #             cur_max = max_selected_subresult_index[data.parent_index]
    #             max_selected_subresult_index[data.parent_index] = max(cur_max, data.index)

    #     selected_count = 0
    #     for index in selected_results:
    #         max_selected_subresult_index.pop(index, None)
    #         selected_count += self._current_results[index].count

    #     partial_selections = []
    #     for parent_index, subresult_index in max_selected_subresult_index.items():
    #         parent = self._current_results[parent_index]
    #         count = parent.subresult(subresult_index).count
    #         selected_count += count
    #         partial_selections.append((parent.text, count, parent.count))

    #     if self._current_results is self._league_results:
    #         total = self._total_league_results
    #     else:
    #         total = self._total_daily_results

    #     text = [f'<strong>Total:</strong> {selected_count / total * 100:0.2f}% ({selected_count}/{total})']
    #     for partial_text, partial_count, partial_total in partial_selections:
    #         text.append(f'<strong>{partial_text}:</strong> {partial_count / partial_total * 100:0.2f}% '
    #                     f'({partial_count}/{partial_total})')
    #     text = '<br />'.join(text)
    #     self.lbl_selected_stats.setText(text)

    def set_price_check_visible(self, visible: bool):
        self.btn_price_check.setVisible(visible)

    def set_results(self, title, league_results: Optional[Tuple[int, List[DisplayResult]]],
                    daily_results: Optional[Tuple[int, List[DisplayResult]]], results_str: str,
                    selection_changed_handler: Callable):
        """Sets the results for this widget.

        Args:
            league_results: the results from the league scrape (possibly an empty list) or `None` if
              league_results are not to be shown
            daily_results: the results from the daily scrape (possibly an empty list) or `None` if
              daily_results are not to be shown
        """
        self._results_str = results_str

        self.lbl_title.setText(title)

        if league_results is not None and daily_results is not None:
            self.toggle_type.show()

        if league_results is not None:
            count, self._league_results = league_results
            self.lbl_league.setText(_LEAGUE_FORMAT.format(count))
            self.lbl_league.show()

        if daily_results is not None:
            count, self._daily_results = daily_results
            self.lbl_daily.setText(_DAILY_FORMAT.format(count))
            self.lbl_daily.show()

        results_league, results_daily = self._build_results(self._league_results, self._daily_results, selection_changed_handler=selection_changed_handler)
        self._widget_league_results = results_league
        self._widget_daily_results = results_daily

        if results_league is not None:
            self.stack_results.setCurrentWidget(results_league)
            self._current_results = self._league_results
            self._set_active_type(is_daily=False)
        else:
            self.stack_results.setCurrentWidget(self._widget_daily_results)
            self._current_results = self._daily_results
            self._set_active_type(is_daily=True)

    def _build_results(
        self,
        *results_lists: List[Optional[List[DisplayResult]]],
        selection_changed_handler: Callable = None
    ) -> List[Optional[QtWidgets.QListWidget]]:
        if not results_lists:
            return None

        widgets = []
        for results in results_lists:
            if results is None:
                widgets.append(None)
            else:
                widget = self._build_result_list(results)
                widgets.append(widget)
                self.stack_results.addWidget(widget)
                if selection_changed_handler:
                    self._connect_signal_to_slot(widget.itemSelectionChanged, selection_changed_handler)

        return widgets

    def _build_result_list(self, results: List[DisplayResult]):
        list_results = QtWidgets.QListWidget(self)
        list_results.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        list_results.setContextMenuPolicy(Qt.CustomContextMenu)

        if results:
            for result in results:
                self._add_result_to_list(result, list_results)
        else:
            item = QtWidgets.QListWidgetItem('Няма резултати')
            item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
            list_results.addItem(item)

        # list_results.itemDoubleClicked.connect(self._on_item_double_clicked)
        list_results.itemSelectionChanged.connect(self._on_results_selection_changed)
        list_results.customContextMenuRequested.connect(self._show_context_menu)

        return list_results

    def _add_result_to_list(self, result: DisplayResult, list_widget: QtWidgets.QListWidget = None,
                            index=None):
        if list_widget is None:
            list_widget = self.stack_results.currentWidget()

        item_result = QtWidgets.QListWidgetItem()
        item_result.setData(Qt.UserRole, result)

        if index is None:
            list_widget.addItem(item_result)
        else:
            list_widget.insertItem(index, item_result)

        widget_result = QtWidgets.QWidget(self)

        lbl_count = QtWidgets.QLabel(str(result.count), widget_result)
        lbl_count.setMinimumWidth(_INDENT)
        lbl_count.setAlignment(Qt.AlignRight)
        lbl_name = QtWidgets.QLabel(result.text, widget_result)

        layout_result = QtWidgets.QHBoxLayout()
        layout_result.setContentsMargins(0, 0, 0, 0)
        if result.indent_level:
            layout_result.addSpacing(result.indent_level * _INDENT)
        layout_result.addWidget(lbl_count, 0, Qt.AlignmentFlag.AlignRight)
        layout_result.addWidget(lbl_name, 1)
        widget_result.setLayout(layout_result)

        widget_result.adjustSize()
        item_result.setSizeHint(widget_result.sizeHint())

        list_widget.setItemWidget(item_result, widget_result)

    def _add_space_to_list(self, list_widget: QtWidgets.QListWidget = None, index=None):
        if list_widget is None:
            list_widget = self.stack_results.currentWidget()

        item = QtWidgets.QListWidgetItem()
        item.setFlags(Qt.NoItemFlags)
        item.setSizeHint(QtCore.QSize(list_widget.sizeHint().width(), 5))

        if index is None:
            list_widget.addItem(item)
        else:
            list_widget.insertItem(index, item)

    def set_selected_stats_text(self, text):
        self.lbl_selected_stats.setText(text)

    def set_price_check_handler(self, handler):
        self._connect_signal_to_slot(self.btn_price_check.clicked, handler)

    def get_selected_data(self):
        return [item.data(Qt.UserRole) for item in self.stack_results.currentWidget().selectedItems()]

    # Properties
    @utils.checkbox_property
    def scrape_type(self):
        return self.toggle_type



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['labbie',
 'labbie.di',
 'labbie.ui',
 'labbie.ui.about',
 'labbie.ui.about.widget',
 'labbie.ui.about.window',
 'labbie.ui.app',
 'labbie.ui.error',
 'labbie.ui.error.widget',
 'labbie.ui.error.window',
 'labbie.ui.result.widget',
 'labbie.ui.screen_selection.widget',
 'labbie.ui.search',
 'labbie.ui.search.widget',
 'labbie.ui.search.window',
 'labbie.ui.settings',
 'labbie.ui.settings.widget',
 'labbie.ui.settings.window',
 'labbie.ui.system_tray',
 'labbie.vendor.qtmodern']

package_data = \
{'': ['*'], 'labbie': ['vendor/qtmodern/resources/*']}

install_requires = \
['Pillow>=8.4.0,<9.0.0',
 'PyQt5>=5.15.6,<6.0.0',
 'aiohttp>=3.8.0,<4.0.0',
 'dacite>=1.6.0,<2.0.0',
 'datrie>=0.8.2,<0.9.0',
 'injector>=0.18.4,<0.19.0',
 'keyboard>=0.13.5,<0.14.0',
 'loguru>=0.5.3,<0.6.0',
 'numpy>=1.21.4,<2.0.0',
 'opencv-python-headless>=4.5.4,<5.0.0',
 'orjson>=3.6.4,<4.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'pytesseract>=0.3.8,<0.4.0',
 'qasync>=0.22.0,<0.23.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['labbie = labbie.__main__:main',
                     'package = package:package']}

setup_kwargs = {
    'name': 'labbie',
    'version': '0.6.1',
    'description': 'Lab enchant assistant',
    'long_description': None,
    'author': 'Brandon Norick',
    'author_email': 'b.norick@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



import atexit
from typing import Dict, List, Optional

from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from labbie.ui import base
from labbie.ui import checkable_combo
from labbie.ui import fuzzy_combo
from labbie.ui import utils

Qt = QtCore.Qt


class TabBar(QtWidgets.QTabBar):

    middleClicked = QtCore.pyqtSignal(int)

    def __init__(self):
        super(QtWidgets.QTabBar, self).__init__()
        self.previousMiddleIndex = -1

    def mousePressEvent(self, mouseEvent):
        if mouseEvent.button() == Qt.MidButton:
            self.previousIndex = self.tabAt(mouseEvent.pos())
        QtWidgets.QTabBar.mousePressEvent(self, mouseEvent)

    def mouseReleaseEvent(self, mouseEvent):
        if mouseEvent.button() == Qt.MidButton and \
            self.previousIndex == self.tabAt(mouseEvent.pos()):
            self.middleClicked.emit(self.previousIndex)
        self.previousIndex = -1
        QtWidgets.QTabBar.mouseReleaseEvent(self, mouseEvent)


class SearchWidget(base.BaseWidget):

    def __init__(self, parent=None):
        super().__init__(parent)

        self._position_path = None
        self._position = None

        lbl_mod = QtWidgets.QLabel('Enchant', self)
        self.combo_mod = fuzzy_combo.FuzzyComboBox(self)
        self.btn_search_mod = QtWidgets.QPushButton('Търсене', self)

        lbl_base = QtWidgets.QLabel('Base', self)
        self.combo_base = fuzzy_combo.FuzzyComboBox(self)
        lbl_ilvl = QtWidgets.QLabel('ilvl', self)
        self.edit_ilvl = QtWidgets.QLineEdit(self)
        lbl_influence = QtWidgets.QLabel('Influence', self)
        self.chkcombo_influences = checkable_combo.CheckableComboBox(self, max_checked=2)
        self.btn_search_base = QtWidgets.QPushButton('Търсене', self)

        self.btn_all = QtWidgets.QPushButton('Всички', self)
        self.btn_screen_capture = QtWidgets.QPushButton('От Екрана', self)

        self.tabs = QtWidgets.QTabWidget(self)
        self.tab_bar = TabBar()
        self.tabs.setTabBar(self.tab_bar)
        self.tab_bar.middleClicked.connect(self.on_tab_middle_click)

        self.combo_mod.setEditable(True)
        self.combo_base.setEditable(True)

        layout_mod_section = QtWidgets.QVBoxLayout()

        layout_mod = QtWidgets.QHBoxLayout()
        layout_mod.addWidget(lbl_mod)
        layout_mod.addWidget(self.combo_mod, 1)
        layout_mod.addWidget(self.btn_search_mod)

        layout_mod_section.addLayout(layout_mod)

        layout_base_section = QtWidgets.QVBoxLayout()

        layout_base = QtWidgets.QHBoxLayout()
        layout_base.addWidget(lbl_base)
        layout_base.addWidget(self.combo_base, 1)
        layout_base.addWidget(lbl_ilvl)
        layout_base.addWidget(self.edit_ilvl)
        layout_base.addWidget(lbl_influence)
        layout_base.addWidget(self.chkcombo_influences)
        layout_base.addWidget(self.btn_search_base)

        layout_base_section.addLayout(layout_base)

        layout_screen_capture = QtWidgets.QHBoxLayout()
        layout_screen_capture.addWidget(self.btn_all)
        layout_screen_capture.addWidget(self.btn_screen_capture, 1)

        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(layout_mod_section)
        layout.addSpacing(10)
        layout.addLayout(layout_base_section)
        layout.addSpacing(10)
        layout.addLayout(layout_screen_capture)
        layout.addSpacing(10)
        layout.addWidget(self.tabs)
        self.setLayout(layout)

        self.setWindowTitle('Търсене на Enchants')
        utils.register_exit_handler(self._at_exit)

    def on_tab_middle_click(self, index):
        self.tabs.removeTab(index)

    def get_position(self):
        pos = self.mapToGlobal(self.pos())
        return pos.x(), pos.y()

    def set_position(self, position):
        if not position:
            self.center_on_screen(adjust_size=False)
        else:
            self.move(*position)

    def set_position_path(self, path):
        self._position_path = path

    def set_search_mod_handler(self, handler):
        self._connect_signal_to_slot(self.btn_search_mod.clicked, handler)

    def set_search_base_handler(self, handler):
        self._connect_signal_to_slot(self.btn_search_base.clicked, handler)

    def set_all_handler(self, handler):
        self._connect_signal_to_slot(self.btn_all.clicked, handler)

    def set_screen_capture_handler(self, handler):
        self._connect_signal_to_slot(self.btn_screen_capture.clicked, handler)

    def set_influence_options(self, influences, data):
        self.chkcombo_influences.addItems(influences, data)

    def set_mods(self, mods: List[str], equivalent_mods: Optional[Dict[str, str]]):
        if equivalent_mods:
            prev_selected = equivalent_mods.get(self.combo_mod.currentText())
        else:
            prev_selected = self.combo_mod.currentText()

        self.combo_mod.clear()
        selected = False
        for index, mod in enumerate(mods):
            select = (prev_selected == mod)
            selected |= select
            self.combo_mod.addItem(mod)
            if select:
                self.combo_mod.setCurrentIndex(index)

        if not selected:
            self.combo_mod.setCurrentIndex(0)

    def set_selected_mod(self, text):
        self.combo_mod.setCurrentText(text)

    def set_bases(self, bases: List[str]):
        prev_selected = self.combo_base.currentText()
        self.combo_base.clear()
        selected = False
        for index, base in enumerate(bases):
            select = (prev_selected == base)
            selected |= select
            self.combo_base.addItem(base)
            if select:
                self.combo_base.setCurrentIndex(index)

        if not selected:
            self.combo_base.setCurrentIndex(0)

    def add_result_tab(self, title, widget: QtWidgets.QWidget, switch=False):
        index = self.tabs.addTab(widget, title)
        if switch:
            self.tabs.setCurrentIndex(index)

    def clear_results(self):
        for _ in range(self.tabs.count()):
            self.tabs.removeTab(0)

    def _at_exit(self):
        if self._position_path:
            with self._position_path.open('w', encoding='utf8') as f:
                x, y = self.get_position()
                f.write(f'{x} {y}')

    # Properties
    @utils.combo_box_property
    def mod(self):
        return self.combo_mod

    @utils.combo_box_property
    def base(self):
        return self.combo_base

    @utils.text_property
    def ilvl(self):
        return self.edit_ilvl

    @utils.checkable_combo_box_property
    def influences(self):
        return self.chkcombo_influences



from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5 import QtWidgets


class ClickableLabel(QtWidgets.QLabel):

    clicked = QtCore.pyqtSignal()

    def mousePressEvent(self, ev: QtGui.QMouseEvent) -> None:
        self.clicked.emit()



import dataclasses
import functools

import injector

from labbie import resources
from labbie import utils


@dataclasses.dataclass
class _Helm:
    display_text: str
    base: str
    unique: bool
Helm = utils.make_slotted_dataclass(_Helm)  # noqa: E305


@injector.singleton
class Bases:

    @injector.inject
    def __init__(self, resource_manager: resources.ResourceManager):
        self._items = resource_manager.items

        self.helms = self._build_helms()

    @functools.cached_property
    def helm_display_texts(self):
        return sorted(helm.display_text for helm in self.helms.values())

    def _build_helms(self):
        helms = {}
        for unique, display_text, base in self._items['helmet']:
            helms[display_text] = Helm(display_text=display_text, base=base, unique=unique)
        return helms
